# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: ecnt_statistics.py
"""
import warnings
import numpy as np
import math

from metcalcpy.util.utils import round_half_up, sum_column_data_by_name, PRECISION, get_total_values

def calculate_ecnt_crps(input_data, columns_names, aggregation=False):
    """Performs calculation of ECNT_CRPS - The Continuous Ranked Probability Score

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ECNT_CRPS as float
            or None if some data values are missing or invalid
    """

    return weighted_average(input_data, columns_names, 'crps', aggregation)


def calculate_ecnt_crpscl(input_data, columns_names, aggregation=False):
    """Performs calculation of ECNT_CRPSCL - Climatological Continuous Ranked Probability Score
        (normal distribution)

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ECNT_CRPSCL as float
            or None if some data values are missing or invalid
    """
    return weighted_average(input_data, columns_names, 'crpscl', aggregation)


def calculate_ecnt_crpss(input_data, columns_names, aggregation=False):
    """Performs calculation of ECNT_CRPSS - The Continuous Ranked Probability Skill Score
        (normal distribution)

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ECNT_CRPSS as float
            or None if some data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        crpscl = sum_column_data_by_name(input_data, columns_names, 'crpscl') / total
        crps = sum_column_data_by_name(input_data, columns_names, 'crps') / total
        crpss = 1 - crps / crpscl
        result = round_half_up(crpss, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result

def calculate_ecnt_crps_emp(input_data, columns_names, aggregation=False):
    """Performs calculation of ECNT_CRPS_EMP - The Continuous Ranked Probability Score
        (empirical distribution)

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ECNT_CRPS_EMP as float
            or None if some data values are missing or invalid
    """

    return weighted_average(input_data, columns_names, 'crps_emp', aggregation)


def calculate_ecnt_crps_emp_fair(input_data, columns_names, aggregation=False):
    """Performs calculation of ECNT_CRPS_EMP_FAIR - The Continuous Ranked Probability Score
        (empirical distribution) adjusted by the mean absolute difference of the ensemble members 

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ECNT_CRPS_EMP_FAIR as float
            or None if some data values are missing or invalid
    """

    return weighted_average(input_data, columns_names, 'crps_emp_fair', aggregation)


def calculate_ecnt_spread_md(input_data, columns_names, aggregation=False):
    """Performs calculation of SPREAD_MD - The pairwise Mean Absolute Difference
        of the unperturbed ensemble members

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated SPREAD_MD as float
            or None if some data values are missing or invalid
    """
    return weighted_average(input_data, columns_names, 'spread_md', aggregation)


def weighted_average(input_data, columns_names, column_name, aggregation=False):
    """ Performs aggregation over multiple cases using a weighted average approach,
     where the weight is defined by the number of matched pairs in the TOTAL column

    :param input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
    :param columns_names: names of the columns for the 2nd dimension as Numpy array
    :param column_name: name of the column to be aggregated
    :param aggregation: if the aggregation on fields was performed
    :return: aggregated column values or None if some data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        statistic = sum_column_data_by_name(input_data, columns_names, column_name) / total
        result = round_half_up(statistic, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_ecnt_crpscl_emp(input_data, columns_names, aggregation=False):
    """Performs calculation of ECNT_CRPSCL_EMP - Climatological Continuous Ranked Probability Score
        (empirical distribution)

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ECNT_CRPSCL_EMP as float
            or None if some data values are missing or invalid
    """

    return weighted_average(input_data, columns_names, 'crpscl_emp', aggregation)

def calculate_ecnt_crpss_emp(input_data, columns_names, aggregation=False):
    """Performs calculation of ECNT_CRPSS_EMP - The Continuous Ranked Probability Skill Score
        (empirical distribution)

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ECNT_CRPSS_EMP as float
            or None if some data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        crps_emp = sum_column_data_by_name(input_data, columns_names, 'crps_emp') / total
        crpscl_emp = sum_column_data_by_name(input_data, columns_names, 'crpscl_emp') / total
        crpss_emp =  1 - crps_emp/crpscl_emp
        result = round_half_up(crpss_emp, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_ecnt_ign(input_data, columns_names, aggregation=False):
    """Performs calculation of ECNT_IGN - The Ignorance Score

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ECNT_IGN as float
            or None if some data values are missing or invalid
    """

    return weighted_average(input_data, columns_names, 'ign', aggregation)


def calculate_ecnt_me(input_data, columns_names, aggregation=False):
    """Performs calculation of ECNT_ME - The Mean Error of the ensemble mean
    (unperturbed or supplied)

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ECNT_ME as float
            or None if some data values are missing or invalid
    """

    return weighted_average(input_data, columns_names, 'me', aggregation)

def calculate_ecnt_mae(input_data, columns_names, aggregation=False):
    """Performs calculation of ECNT_MAE - The Mean Absolute Error of the ensemble mean
        (unperturbed or supplied)

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ECNT_MAE as float
            or None if some data values are missing or invalid
    """

    return weighted_average(input_data, columns_names, 'mae', aggregation)

def calculate_ecnt_mae_oerr(input_data, columns_names, aggregation=False):
    """Performs calculation of MAE_OERR - The Mean Absolute Error of the PERTURBED
        ensemble mean (e.g. with Observation Error)

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated MAE_OERR as float
            or None if some data values are missing or invalid
    """

    return weighted_average(input_data, columns_names, 'mae_oerr', aggregation)


def calculate_ecnt_rmse(input_data, columns_names, aggregation=False):
    """Performs calculation of ECNT_RMSE - The Root Mean Square Error of the ensemble mean
    (unperturbed or supplied)

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ECNT_RMSE as float
            or None if some data values are missing or invalid
    """

    warnings.filterwarnings('error')
    try:
        wa = weighted_average(input_data, columns_names, 'mse', aggregation)
        rmse = math.sqrt(wa)
        result = round_half_up(rmse, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_ecnt_spread(input_data, columns_names, aggregation=False):
    """Performs calculation of ECNT_SPREAD - The mean of the spread (standard deviation)
        of the unperturbed ensemble member values at each observation location

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ECNT_SPREAD as float
            or None if some data values are missing or invalid
    """

    warnings.filterwarnings('error')
    try:
        wa = weighted_average(input_data, columns_names, 'variance', aggregation)
        spread = math.sqrt(wa)
        result = round_half_up(spread, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_ecnt_me_oerr(input_data, columns_names, aggregation=False):
    """Performs calculation of ECNT_ME_OERR - The Mean Error of the PERTURBED ensemble mean
        (e.g. with Observation Error)
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ECNT_ME_OERR as float
            or None if some data values are missing or invalid
    """

    return weighted_average(input_data, columns_names, 'me_oerr', aggregation)


def calculate_ecnt_rmse_oerr(input_data, columns_names, aggregation=False):
    """Performs calculation of ECNT_RMSE_OERR - The Root Mean Square Error of the PERTURBED
        ensemble mean (e.g.with Observation Error)
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ECNT_RMSE_OERR as float
            or None if some data values are missing or invalid
    """

    warnings.filterwarnings('error')
    try:
        wa = weighted_average(input_data, columns_names, 'mse_oerr', aggregation)
        mse_oerr = math.sqrt(wa)
        result = round_half_up(mse_oerr, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_ecnt_spread_oerr(input_data, columns_names, aggregation=False):
    """Performs calculation of ECNT_SPREAD_OERR - The mean of the spread (standard deviation)
        of the PERTURBED ensemble member values (e.g. with Observation Error )
        at each observation location

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ECNT_SPREAD_OERR as float
            or None if some data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        wa = weighted_average(input_data, columns_names, 'variance_oerr', aggregation)
        spread_oerr = math.sqrt(wa)
        result = round_half_up(spread_oerr, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_ecnt_spread_plus_oerr(input_data, columns_names, aggregation=False):
    """Performs calculation of SPREAD_PLUS_OERR - The square root of the sum of
        unperturbed ensemble variance and the observation error variance
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated SPREAD_PLUS_OERR as float
            or None if some data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        wa = weighted_average(input_data, columns_names, 'variance_plus_oerr', aggregation)
        spread_plus_oerr = math.sqrt(wa)
        result = round_half_up(spread_plus_oerr, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result

def calculate_ecnt_n_ge_obs(input_data, columns_names, aggregation=False):
    """Performs calculation of N_GE_OBS - The number of ensemble values greater
        than or equal to their observations
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated N_GE_OBS as float
            or None if some data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        n_ge_obs = sum_column_data_by_name(input_data, columns_names, 'n_ge_obs')
        result = round_half_up(n_ge_obs, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result

def calculate_ecnt_n_lt_obs(input_data, columns_names, aggregation=False):
    """Performs calculation of N_LT_OBS - The number of ensemble values less
        than their observations
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated N_LT_OBS as float
            or None if some data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        n_lt_obs = sum_column_data_by_name(input_data, columns_names, 'n_lt_obs')
        result = round_half_up(n_lt_obs, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result

def calculate_ecnt_me_ge_obs(input_data, columns_names, aggregation=False):
    """Performs calculation of ME_GE_OBS - The Mean Error of the ensemble values
        greater than or equal to their observations
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ME_GE_OBS as float
            or None if some data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        n_ge_obs = sum_column_data_by_name(input_data, columns_names, 'n_ge_obs')
        me_ge_obs = sum_column_data_by_name(input_data, columns_names, 'me_ge_obs')/n_ge_obs
        result = round_half_up(me_ge_obs, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result

def calculate_ecnt_me_lt_obs(input_data, columns_names, aggregation=False):
    """Performs calculation of ME_GE_OBS - The Mean Error of the ensemble values
        greater than or equal to their observations
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ME_GE_OBS as float
            or None if some data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        n_lt_obs = sum_column_data_by_name(input_data, columns_names, 'n_lt_obs')
        me_lt_obs = sum_column_data_by_name(input_data, columns_names, 'me_lt_obs')/n_lt_obs
        result = round_half_up(me_lt_obs, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result

def calculate_ecnt_bias_ratio(input_data, columns_names, aggregation=False):
    """Performs calculation of BIAS_RATIO - The Bias Ratio
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated BIAS_RATIO as float
            or None if some data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        me_ge_obs = calculate_ecnt_me_ge_obs(input_data, columns_names)
        me_lt_obs = calculate_ecnt_me_lt_obs(input_data, columns_names)
        bias_ratio = me_ge_obs/abs(me_lt_obs)
        result = round_half_up(bias_ratio, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_ecnt_total(input_data, columns_names):
    """Performs calculation of Total number of matched pairs for Ensemble Continuous Statistics
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated Total number of matched pairs as float
            or None if some data values are missing or invalid
    """
    total = sum_column_data_by_name(input_data, columns_names, 'total')
    return round_half_up(total, PRECISION)



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: met_stats.py
"""
import math
import numpy as np

__author__ = 'Tatiana Burek'
__version__ = '0.1.0'


def get_column_index_by_name(columns, column_name):
    """Finds the index of the specified column in the array

            Args:
                columns: names of the columns as Numpy array
                column_name: the name of the column

            Returns:
                the index of the column
                or None if the column name does not exist in the array
    """
    index_array = np.where(columns == column_name)[0]
    if index_array.size == 0:
        return None
    return index_array[0]


def calc_direction(u_comp, v_comp):
    """ Calculated the direction of the wind from it's u and v components in degrees
        Args:
            u_comp: u wind component
            v: v wind component

        Returns:
            direction of the wind in degrees or None if one of the components is less then tolerance
    """
    tolerance = 1e-5
    if abs(u_comp) < tolerance and abs(v_comp) < tolerance:
        return None

    direction = np.arctan2(u_comp, v_comp)
    # convert to [0,360]
    direction = direction - 360 * math.floor(direction / 360)
    return direction


def calc_speed(u_comp, v_comp):
    """ Calculated the speed of the wind from it's u and v components
        Args:
            u_comp: u wind component
            v_comp: v wind component

        Returns:
            speed of the wind  or None
    """
    try:
        result = np.sqrt(u_comp * u_comp + v_comp * v_comp)
    except (TypeError, Warning):
        result = None
    return result



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: rps_statistics.py
"""
import warnings

from metcalcpy.util.utils import round_half_up, sum_column_data_by_name, PRECISION, get_total_values

__author__ = 'Tatiana Burek'
__version__ = '0.1.0'


def calculate_rps(input_data, columns_names, aggregation=False):
    """Performs calculation of RPS - Ranked Probability Score

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated RPS as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        rps = sum_column_data_by_name(input_data, columns_names, 'rps') / total
        result = round_half_up(rps, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_rps_comp(input_data, columns_names, aggregation=False):
    """Performs calculation of RPS_COMP - Complement of the Ranked Probability Score
       It is computed simply as RPS_COMP = 1 - RPS

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated RPS as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        rps_comp = sum_column_data_by_name(input_data, columns_names, 'rps_comp') / total
        result = round_half_up(rps_comp, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_rpss(input_data, columns_names, aggregation=False):
    """Performs calculation of RPSS -

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated RPS as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        rps = sum_column_data_by_name(input_data, columns_names, 'rps') / total
        rps_climo = sum_column_data_by_name(input_data, columns_names, 'rps_climo') / total
        rpss = 1 - rps / rps_climo
        result = round_half_up(rpss, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_rps_total(input_data, columns_names):
    """Performs calculation of Total number of matched pairs for
        Ranked Probability Score Statistics
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated Total number of matched pairs as float
            or None if some of the data values are missing or invalid
    """
    total = sum_column_data_by_name(input_data, columns_names, 'total')
    return round_half_up(total, PRECISION)



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
import numpy as np
import xarray as xr
"""
local scripts, if loading from a different directory include that with a '.' between
directory name and script name
"""
import ccew_activity as ccew

"""
Paths to plot and data directories. The annual EOF files for the waves are provided. The eofpath should point 
to the location of these files. The filenames need to match what is being read in in ccew_activity.waveact
"""
# plotpath = '../plots/'
# eofpath = '../data/EOF/'
# datapath = '../data/'
plotpath = '/Users/fillmore/working/'
eofpath = '/Volumes/d1/fillmore/Data/TropicalDiagnostics/'
datapath = '/Volumes/d1/fillmore/Data/TropicalDiagnostics/'
"""
Parameters to set for plotting Kelvin activity index.
"""
wave = 'Kelvin'
datestrt = '2015-12-01 00:00:00'
datelast = '2016-03-31 13:00:00'


print("reading ERAI data from file:")
# spd = 1
spd = 2
ds = xr.open_dataset(datapath+'/precip.erai.sfc.1p0.'+str(spd)+'x.2014-2016.nc')
A = ds.precip
print("extracting time period:")
A = A.sel(time=slice(datestrt, datelast))
A = A.squeeze()
timeA = ds.time.sel(time=slice(datestrt, datelast))
ds.close()
A = A * 1000/4
A.attrs['units'] = 'mm/d'

print("project data onto wave EOFs")
waveactA = ccew.waveact(A, wave, eofpath, spd, '1p0', 180, 'annual')
print(waveactA.min(), waveactA.max())


print("reading observed precipitation data from file:")
# spd = 1
spd = 2
ds = xr.open_dataset(datapath+'/precip.trmm.'+str(spd)+'x.1p0.v7a.fillmiss.comp.2014-2016.nc')
B = ds.precip
print("extracting time period:")
B = B.sel(time=slice(datestrt, datelast))
B = B.squeeze()
timeB = ds.time.sel(time=slice(datestrt, datelast))
ds.close()
B.attrs['units'] = 'mm/d'

print("project data onto wave EOFs")
waveactB = ccew.waveact(B, wave, eofpath, spd, '1p0', 180, 'annual')
print(waveactB.min(), waveactB.max())

exps = [0, 1]
explabels = ['trmm', 'erai']
nexps = len(exps)

print("computing skill")
skill = ccew.wave_skill(B)


##### maybe this next routine needs to be moved to METplotpy? ############
# ccew.plot_skill(skill, wave, explabels, plotpath)



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: test_validate_mv_python.py


How to use:
    python test_validate_mv_python.py <parameters_file>
        where - <parameters_file> is YAML file with parameters
        and environment variable should be set to PYTHONPATH=<path_to_METcalcpy>

    example.yaml:
    input_xml_dir:      /path/to/dir_with_xml_files
    output_xml_dir:     /path/to/dir_with_new_xml_files
    output_plots_dir:   /path/to/dir_with_new_plots_files
    r_dir:              /path/to/dir_with_r_plots
    mv_batch:           /path/to/metviewer_batch_script
    db_user:            db_user_name
    db_password:        db_pass
    start_date:         Optional. Start date of testing %Y-%m-%d. If not specified = yesterday
    end_date:           Optional. End date of testing %Y-%m-%d. If not specified = today


"""
import argparse
import glob
import os
import datetime as dt
import sys
import xml.dom.minidom
import subprocess
import yaml

from metcalcpy.compare_images import CompareImages


def replace_name(old_name, postfix):
    """Adds postfix to the end of a file name but before the extension

       Args:
           old_name: file name to be changed
           postfix: postfix
       Returns:
           file name with postfix
    """
    return old_name.replace(".", "_" + postfix + ".")


def clean(params):
    """Removes all files form the previous run

        Args:
            params: input parameters as a dictionary
        """

    for file in os.scandir(params['output_xml_dir']):
        os.unlink(file.path)

    for file in os.scandir(params['output_plots_dir']):
        os.unlink(file.path)

    for file in os.scandir(params['output_data_dir']):
        os.unlink(file.path)
    for file in os.scandir(params['output_scripts_dir']):
        os.unlink(file.path)


def main(params):
    """Finds METViewer generated xml files based on the values from the parameters dictionary.
        Adjust XML and reruns it using the Python version of METViewer
        Compares the original and generated images and prints the result

        Args:
            params: input parameters as a dictionary

    """
    # remove old output
    clean(params)

    # find XML files
    test_xml = get_test_xml(params)

    # rerun each XML with Python
    for file in test_xml:
        print(f'\nChecking {file}')
        doc = xml.dom.minidom.parse(file)

        plot_name = doc.getElementsByTagName('plot_file')[0].firstChild.nodeValue
        plot_path = doc.getElementsByTagName('plots')[0].firstChild.nodeValue
        original_plot_path = plot_path + '/' + plot_name

        # replace the original XML with new values
        doc.getElementsByTagName('plot_file')[0].firstChild.nodeValue \
            = replace_name(plot_name, 'py')
        doc.getElementsByTagName('data_file')[0].firstChild.nodeValue \
            = replace_name(doc.getElementsByTagName('data_file')[0].firstChild.nodeValue, 'py')
        doc.getElementsByTagName('r_file')[0].firstChild.nodeValue \
            = replace_name(doc.getElementsByTagName('r_file')[0].firstChild.nodeValue, 'py')
        doc.getElementsByTagName('r_tmpl')[0].firstChild.nodeValue = params['mv_home'] + 'R_tmpl'
        doc.getElementsByTagName('r_work')[0].firstChild.nodeValue = params['mv_home'] + 'R_work'
        doc.getElementsByTagName('plots')[0].firstChild.nodeValue = params['output_plots_dir']
        doc.getElementsByTagName('user')[0].firstChild.nodeValue = params['db_user']
        doc.getElementsByTagName('password')[0].firstChild.nodeValue = params['db_pass']
        doc.getElementsByTagName('data')[0].firstChild.nodeValue = params['output_data_dir']
        doc.getElementsByTagName('scripts')[0].firstChild.nodeValue = params['output_scripts_dir']

        # save new XML
        new_xml_file = params['output_xml_dir'] + os.path.basename(replace_name(file, 'py'))
        with open(new_xml_file, "w") as xml_file:
            doc.writexml(xml_file)

        # run METviewer with the new XML and wait till it is done
        process = subprocess.Popen([params['mv_home'] + '/bin/mv_batch.sh', new_xml_file],
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True)

        (output, err) = process.communicate()
        p_status = process.wait()

        new_image_path = params['output_plots_dir'] + replace_name(plot_name, 'py')

        # check if both images are present or not
        if not os.path.exists(original_plot_path) \
                and not os.path.exists(new_image_path):
            # if both images don't exist - success
            print(f'SUCCESS: For {plot_name} both images don\'t exist')
            # remove new XML
            os.remove(new_xml_file)
            # remove new script
            delete_similar_files(params['output_scripts_dir'], plot_name)
            # remove new data files
            delete_similar_files(params['output_data_dir'], plot_name)

        else:
            # compare images
            try:
                compare = CompareImages(original_plot_path, new_image_path)
                ssim = compare.get_mssim()
                if ssim == 1.0:
                    print(f'SUCCESS: For {plot_name} images are identical')
                    # remove new image
                    os.remove(new_image_path)
                    # remove new XML
                    os.remove(new_xml_file)
                    # remove new script
                    delete_similar_files(params['output_scripts_dir'], plot_name)
                    # remove new data files
                    delete_similar_files(params['output_data_dir'], plot_name)

                else:
                    print(f'ERROR: For {plot_name} images are different')
                    # add more diagnostic images
                    compare.save_thresh_image(params['output_plots_dir']
                                              + replace_name(plot_name, 'thresh'))

            except KeyError as err:
                print(f'ERROR: For {plot_name} : {err}')


def delete_similar_files(directory, file_name):
    """ Deletes all files that have the same name but different extensions from the directory

        Args:
            directory - directory name where to look for files
            file_name - name of the file to delete

    """
    for filename in glob.glob(directory
                              + file_name.split('.')[0]
                              + '*'):
        os.remove(filename)


def get_test_xml(params):
    """Creates a list of xml fies based on the parameters values

       Args:
           params: input parameters as a dictionary that include required key'input_xml_dir'
           and optional 'start_date' and 'end_date'
       Returns:
           a list with xml file names
    """
    # calculate testing period
    (start, end) = get_testing_period(params)
    print(f'Tested period {start} - {end}')

    fresh_xml = []
    # for each fie in the directory check the date and extension
    # and include it to the output list if it fits the criteria
    for file in os.listdir(params['input_xml_dir']):
        file_name = params['input_xml_dir'] + file
        filetime = dt.datetime.fromtimestamp(os.path.getmtime(file_name))
        extension = os.path.splitext(file)[1]
        if start <= filetime.date() <= end and extension == '.xml':
            fresh_xml.append(file_name)

    fresh_xml.sort()
    return fresh_xml


def get_testing_period(params):
    """Using parameters forom the input dictionary creates the date range.
        If 'start_date' parameter is not present initialises it as yesterday date
        If 'end_date' parameter is not present initialises it as today date

        Args:
           params: input parameters as a dictionary that might include optional keys 'start_date'
           and 'end_date'
        Returns:
           a date range as start and end dates
        Raises: KeyError if the date range is invalid
    """
    if 'start_date' in params.keys():
        start = dt.datetime.strptime(str(params['start_date']), '%Y-%m-%d').date()
    else:
        # start = yesterday
        start = (dt.datetime.now() - dt.timedelta(1)).date()

    if 'end_date' in params.keys():
        end = dt.datetime.strptime(str(params['end_date']), '%Y-%m-%d').date()
    else:
        # end = today
        end = dt.datetime.now().date()

    if end < start:
        raise KeyError(f'Invalid start/end dates {start} - {end}')

    return start, end


if __name__ == "__main__":
    # construct the argument parse and parse the arguments
    PARSER = argparse.ArgumentParser(description='List of validate_mv_python arguments')
    PARSER.add_argument("parameter_file", help="Path to YAML parameters file",
                        type=argparse.FileType('r'),
                        default=sys.stdin)
    ARGS = PARSER.parse_args()
    PARAMS = yaml.load(ARGS.parameter_file, Loader=yaml.FullLoader)
    main(PARAMS)



# ============================*
# ** Copyright UCAR (c) 2022
# ** University Corporation for Atmospheric Research (UCAR)
# ** National Center for Atmospheric Research (NCAR)
# ** Research Applications Lab (RAL)
# ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
# ============================*


"""
Program Name: scorecard.py

How to use:
 - Call from other Python function
        SCORECARD = Scorecard(PARAMS)
        SCORECARD.calculate_scorecard_data()
        where PARAMS – a dictionary with data description parameters including
        location of input and output data.
        The structure is similar to Rscript template

 - Run as a stand-alone script
        python scorecard.py <parameters_file>
        where - <parameters_file> is YAML file with parameters
        and environment variable should be set to PYTHONPATH=<path_to_METcalcpy>

 - Run from Java
        proc = Runtime.getRuntime().exec(
                “python scorecard.py <parameters_file>”,
                new String[]{”PYTHONPATH=<path_to_METcalcpy>”},
                new File(System.getProperty("user.home")));

"""

from typing import Union
import pandas as pd
import yaml
import logging
import argparse
import sys
import itertools
import re
import statistics
import warnings
import math
import numpy as np
from pandas import DataFrame
from pandas.core.common import SettingWithCopyWarning

from metcalcpy import GROUP_SEPARATOR, DATE_TIME_REGEX
from metcalcpy.util.tost_paired import pt

from metcalcpy.util.utils import intersection, get_derived_curve_name, \
    is_derived_point, is_string_integer, OPERATION_TO_SIGN, calc_derived_curve_value, \
    perfect_score_adjustment, sort_data, PRECISION, DerivedCurveComponent, is_string_strictly_float

COLUMNS_TO_REMOVE = ['equalize', 'stat_ncl', 'stat_ncu', 'stat_bcl', 'stat_bcu', 'fcst_valid_beg', 'fcst_init_beg']


class Scorecard:
    """A class that performs calculation requested statistics for the scorecard. Possible statistic values:
              DIFF_SIG - the difference significance
              DIFF     - mean of difference
              SINGLE   - mean value
           Saves the data into the file
        Usage:
            initialise this call with the parameters dictionary and then
            call calculate_stats method
            This method will crate and save to the file aggregation statistics
                scorecard = Scorecard(params)
                Scorecard.calculate_scorecard_data()
        Raises: EmptyDataError or ValueError when the input DataFrame is empty or doesn't have data
    """

    def __init__(self, in_params):
        """Initialises the class by saving input parameters and reading data from file
                Args:
                    in_params - input parameters as a dictionary
                Raises: EmptyDataError or ValueError when the input DataFrame is empty
                    or doesn't have data
        """

        self.params = in_params
        self.derived_name_to_values = {}
        # import pandas
        try:
            self.input_data = pd.read_csv(
                self.params['sum_stat_input'],
                header=[0],
                sep='\t'
            )

            if self.input_data.empty:
                logging.warning('The input data is empty. The empty output file was created')
                export_csv = self.input_data.to_csv(self.params['sum_stat_output'],
                                                    index=None, header=True, mode="w",
                                                    sep="\t")
                raise ValueError("The input data is empty")

            self.column_names = self.input_data.columns.values
        except pd.errors.EmptyDataError:
            raise
        except KeyError as ex:
            logging.error('Parameter with key %s is missing', ex)
            raise
        self.group_to_value = {}

    def calculate_scorecard_data(self):
        """Calculates requested statistics for the scorecard. Possible statistic values:
              DIFF_SIG - the difference significance
              DIFF     - mean of difference
              SINGLE   - mean value
           Saves the data into the file

        """
        # identify all possible points values by adding series values, indy values
        # and statistics and then permute them
        if self.input_data.empty:
            return pd.DataFrame()
        series_val = self.params['series_val_1']

        all_fields_values = series_val.copy()
        indy_vals = self.params['indy_vals']
        all_fields_values['stat_name'] = self.params['list_stat_1']
        if indy_vals:
            all_fields_values[self.params['indy_var']] = indy_vals
        all_points = list(itertools.product(*all_fields_values.values()))
        if self.params['derived_series_1']:
            # identifies and add all possible derived points values
            all_points.extend(self._get_derived_points(series_val, indy_vals))

        point_to_distrib = {}
        derived_frame = None
        # for each point
        for point_ind, point in enumerate(all_points):
            is_derived = is_derived_point(point)
            if not is_derived:
                # only get the data for each point - no calculations needed
                all_filters = []
                all_filters_pct = []
                filters_wihtout_indy = []
                for field_ind, field in enumerate(all_fields_values.keys()):
                    filter_value = point[field_ind]
                    if GROUP_SEPARATOR in filter_value:
                        filter_list = re.findall(DATE_TIME_REGEX, filter_value)
                        if len(filter_list) == 0:
                            filter_list = filter_value.split(GROUP_SEPARATOR)
                    elif ";" in filter_value:
                        filter_list = filter_value.split(';')
                    else:
                        filter_list = [filter_value]
                    for i, filter_val in enumerate(filter_list):
                        if is_string_integer(filter_val):
                            filter_list[i] = int(filter_val)
                        elif is_string_strictly_float(filter_val):
                            filter_list[i] = float(filter_val)
                    if field in self.input_data.keys():
                        if field != self.params['indy_var']:
                            filters_wihtout_indy.append((self.input_data[field].isin(filter_list)))

                        all_filters.append((self.input_data[field].isin(filter_list)))
                    if field in series_val.keys():
                        all_filters_pct.append((self.input_data[field].isin(filter_list)))

                # use numpy to select the rows where any record evaluates to True
                mask = np.array(all_filters).all(axis=0)
                point_data = self.input_data.loc[mask]
                point_to_distrib[point] = point_data

            else:
                # perform statistics calculation
                results = self._get_stats_for_derived(point, point_to_distrib)
                if results is not None:
                    if derived_frame is None:
                        derived_frame = results
                    else:
                        derived_frame = pd.concat([derived_frame, results], axis=0)

        # print the result to file
        if derived_frame is not None:
            header = True
            mode = 'w'
            if 'append_to_file' in self.params.keys() and self.params['append_to_file'] == 'True':
                header = False
                mode = 'a'
            export_csv = derived_frame.to_csv(self.params['sum_stat_output'],
                                          index=None, header=header, mode=mode,
                                          sep="\t", na_rep="NA", float_format='%.' + str(PRECISION) + 'f')

    def _get_stats_for_derived(self, series, series_to_data) -> Union[DataFrame, None]:
        """ Calculates  derived statistic value for input data
            Args:
                series: array of length = 3 where
                1st element - derived series title,
                    ex. 'DIFF(ENS001v3.6.1_d01 DPT FBAR-ENS001v3.6.1_d02 DPT FBAR)'
                others  - additional values like indy val and statistic
                series_to_data - dictionary of the series title to its data as DataFrame

            Returns:
                DataFrame containing 1 row with the resulting data or None

        """

        # get derived name
        derived_name = ''
        for operation in OPERATION_TO_SIGN:
            for point_component in series:
                if point_component.startswith((operation + '(', operation + ' (')):
                    derived_name = point_component
                    break

        # find all components for the 1st and 2nd series
        derived_curve_component = self.derived_name_to_values[derived_name]
        permute_for_first_series = derived_curve_component.first_component.copy()
        for series_comp in series[1:]:
            if series_comp not in permute_for_first_series:
                permute_for_first_series.append(series_comp)

        # replace first_series components group names to values
        for i, perm in enumerate(permute_for_first_series):
            if perm in self.group_to_value:
                permute_for_first_series[i] = self.group_to_value[perm]

        permute_for_second_series = derived_curve_component.second_component.copy()
        for series_comp in series[1:]:
            if series_comp not in permute_for_second_series:
                permute_for_second_series.append(series_comp)

        # replace second_series components group names to values
        for i, perm in enumerate(permute_for_second_series):
            if perm in self.group_to_value:
                permute_for_second_series[i] = self.group_to_value[perm]

        ds_1 = None
        ds_2 = None

        # for each component find  its data as DataFrame
        for series_to_distrib_key in series_to_data.keys():
            if all(elem in permute_for_first_series for elem in series_to_distrib_key):
                ds_1 = series_to_data[series_to_distrib_key]
            if all(elem in permute_for_second_series for elem in series_to_distrib_key):
                ds_2 = series_to_data[series_to_distrib_key]
            if ds_1 is not None and ds_2 is not None:
                break

        if ds_1.values is None or ds_2.values is None \
                or ds_1.values.size == 0 or ds_2.values.size == 0:
            return None

        # validate data
        if derived_curve_component.derived_operation != 'SINGLE' \
                and len(ds_1.values) != 0 and len(ds_2.values):
            fcst_lead_index = np.where(self.column_names == 'fcst_lead')[0][0]
            stat_name_index = np.where(self.column_names == 'stat_name')[0][0]
            if "fcst_valid_beg" in self.column_names:
                fcst_valid_ind = np.where(self.column_names == 'fcst_valid_beg')[0][0]
            elif "fcst_valid" in self.column_names:
                fcst_valid_ind = np.where(self.column_names == 'fcst_valid')[0][0]
            elif "fcst_init_beg" in self.column_names:
                fcst_valid_ind = \
                    np.where(self.column_names == 'fcst_init_beg')[0][0]
            else:
                fcst_valid_ind = \
                    np.where(self.column_names == 'fcst_init')[0][0]
            try:
                # filter columns of interest
                date_lead_stat = ds_1.values[:, [fcst_valid_ind, fcst_lead_index, stat_name_index]]
                # find the number of unique combinations
                unique_date_size = len(set(map(tuple, date_lead_stat)))
            except TypeError as err:
                print(err)
                unique_date_size = []
            if unique_date_size != len(ds_1.values):
                raise NameError("Derived curve can't be calculated."
                                " Multiple values for one valid date/fcst_lead")

        # sort data by dates
        ds_1_values = sort_data(ds_1)
        ds_2_values = sort_data(ds_2)

        stat_values_1 = ds_1_values['stat_value'].tolist()
        stat_values_2 = ds_2_values['stat_value'].tolist()

        # calculate derived statistic based on the operation and stat_flag
        derived_stat = None
        if derived_curve_component.derived_operation == 'DIFF_SIG':
            if self.params['stat_flag'] == 'EMC':
                derived_stat = self._calculate_diff_sig_emc(stat_values_1, stat_values_2)
            else:
                derived_stat = self._calculate_diff_sig_ncar(stat_values_1, stat_values_2)

        elif derived_curve_component.derived_operation == 'DIFF':
            # perform the derived operation
            derived_stat_list = calc_derived_curve_value(
                stat_values_1,
                stat_values_2,
                'DIFF')
            derived_stat = statistics.mean(derived_stat_list)

        elif derived_curve_component.derived_operation == 'SINGLE':
            derived_stat = statistics.mean(stat_values_1)

        # create dataframe from teh 1st row of the original data
        df = ds_1.head(1)
        # remove unneeded columns
        for column in COLUMNS_TO_REMOVE:
            if column in self.column_names:
                df = df.drop(column, axis=1)
        df.reset_index(drop=True, inplace=True)

        # set statistic value
        df.at[0, 'stat_value'] = derived_stat
        df = df.astype({"fcst_lead": str})

        # set fcst_lead
        for fcst_lead in self.params["series_val_1"]['fcst_lead']:
            for component in derived_curve_component.first_component:
                if component == fcst_lead:
                    with warnings.catch_warnings():
                        warnings.simplefilter(action="ignore", category=SettingWithCopyWarning)
                        df.at[0, 'fcst_lead'] = fcst_lead
                    break

        # set model
        df.at[0, 'model'] = derived_name
        return df

    def _calculate_diff_sig_ncar(self, ds_1_values, ds_2_values) -> float:
        """ calculates the difference significance between two data arrays using p-value

            Args:
                    ds_1_values: list of floats
                    ds_2_values: list of floats
            Returns: the difference significance

        """
        # perform the derived operation
        derived_stat_list = calc_derived_curve_value(ds_1_values, ds_2_values, 'DIFF_SIG')
        avg = statistics.mean(derived_stat_list)
        sdv = statistics.stdev(derived_stat_list)
        total = len(derived_stat_list)
        t = avg / (sdv / np.sqrt(total))
        p_val = 1 - 2 * pt(abs(t), total - 1, lower_tail=False)
        derived_stat = perfect_score_adjustment(statistics.mean(ds_1_values),
                                                statistics.mean(ds_2_values),
                                                self.params['list_stat_1'][0],
                                                p_val)
        return derived_stat

    def _calculate_diff_sig_emc(self, ds_1_values, ds_2_values) -> float:
        """ calculates the difference significance between two data arrays using EMC method

            Args:
                    ds_1_values: list of floats
                    ds_2_values: list of floats
            Returns: the difference significance

        """
        derived_stat = None
        values_1 = np.array(ds_1_values)
        values_2 = np.array(ds_2_values)
        val2_minus_val1 = np.subtract(values_2, values_1)
        acdm = sum(val2_minus_val1) / self.params['ndays']
        acdm_list = [acdm] * len(values_1)
        std = math.sqrt(sum(np.subtract(val2_minus_val1, acdm_list) * np.subtract(val2_minus_val1, acdm_list)) /
                        self.params['ndays'])
        nsz = len(ds_1_values)
        intvl = round(1.960 * std / math.sqrt(nsz - 1), 6)
        mean1 = round(statistics.mean(values_1), 6)
        mean2 = round(statistics.mean(values_2), 6)
        if self.params['list_stat_1'][0].startswith('ME') or self.params['list_stat_1'][0].startswith('BIAS'):
            ds = (abs(mean2 - mean1)) / intvl
            sss = abs(mean2) - abs(mean1)
            if sss is not None and sss < 0:
                ds = (-1) * ds
        elif self.params['list_stat_1'][0].startswith('RMSE') \
                or self.params['list_stat_1'][0].startswith('RMSVE'):
            ds = (mean2 - mean1) / intvl
        else:
            ds = (mean1 - mean2) / intvl
        if ds is not None:
            ds = round(ds, 3)
            if self.params['ndays'] >= 80:
                alpha1 = 1.960  # 95% confidence level
                alpha2 = 2.576  # 99% confidence level
                alpha3 = 3.291  # 99.9% confidence level
            elif self.params['ndays'] >= 40 and self.params['ndays'] < 80:
                alpha1 = 2.0  # 95% confidence level
                alpha2 = 2.66  # 99% confidence level
                alpha3 = 3.46  # 99.9% confidence level
            elif self.params['ndays'] >= 20 and self.params['ndays'] < 40:
                alpha1 = 2.042  # 95% confidence level
                alpha2 = 2.75  # 99% confidence level
                alpha3 = 3.646  # 99.9% confidence level
            elif self.params['ndays'] < 20:
                alpha1 = 2.228  # 95% confidence level
                alpha2 = 3.169  # 99% confidence level
                alpha3 = 4.587  # 99.9% confidence level
            ds95 = ds
            ds99 = ds * alpha1 / alpha2
            ds99 = round(ds99, 3)
            ds999 = ds * alpha1 / alpha3;
            ds999 = round(ds999, 3)
            if ds999 >= 1:
                derived_stat = 1
            elif ds999 < 1 and ds99 >= 1:
                derived_stat = 0.99
            elif ds99 < 1 and ds95 >= 1:
                derived_stat = 0.95
            elif ds95 > -1 and ds95 < 1:
                derived_stat = 0
            elif ds95 <= -1 and ds99 > -1:
                derived_stat = -0.95
            elif ds99 <= -1 and ds999 > -1:
                derived_stat = -0.99
            elif ds999 <= -1 and ds999 > -100.0:
                derived_stat = -1
            elif ds999 < -100.0:
                derived_stat = -1
        return derived_stat

    def _get_derived_points(self, series_val, indy_vals):
        """identifies and returns as a list all possible derived points values

            Args:
                series_val: dictionary of series variable to values
                indy_vals: list of independent values
            Returns: a list of all possible values for each derived points

        """

        # for each derived series
        result = []
        for derived_serie in self.params['derived_series_1']:
            # series 1 components
            ds_1 = derived_serie[0].split(' ')

            # series 2 components
            ds_2 = derived_serie[1].split(' ')
            # find a variable of the operation by comparing values in each derived series component
            series_var_vals = ()
            for ind, name in enumerate(ds_1):
                if name != ds_2[ind]:
                    series_var_vals = (name, ds_2[ind])
                    break

            series_var = list(series_val.keys())[-1]
            if len(series_var_vals) > 0:
                for var in series_val.keys():
                    if all(elem in series_val[var] for elem in series_var_vals):
                        series_var = var
                        break

            derived_val = series_val.copy()
            derived_val[series_var] = None

            for var in series_val.keys():
                if derived_val[var] is not None \
                        and intersection(derived_val[var], ds_1) \
                        == intersection(derived_val[var], ds_1):
                    derived_val[var] = intersection(derived_val[var], ds_1)

            derived_curve_name = get_derived_curve_name(derived_serie)
            derived_val[series_var] = [derived_curve_name]
            if len(indy_vals) > 0:
                derived_val[self.params['indy_var']] = indy_vals

            self.derived_name_to_values[derived_curve_name] \
                = DerivedCurveComponent(ds_1, ds_2, derived_serie[-1])
            if ds_1[-1] == ds_2[-1]:
                derived_val['stat_name'] = [ds_1[-1]]
            else:
                derived_val['stat_name'] = [ds_1[-1] + "," + ds_2[-1]]
            result.append(list(itertools.product(*derived_val.values())))

        return [y for x in result for y in x]


if __name__ == "__main__":
    PARSER = argparse.ArgumentParser(description='List of scorecard arguments')
    PARSER.add_argument("parameters_file", help="Path to YAML parameters file",
                        type=argparse.FileType('r'),
                        default=sys.stdin)
    ARGS = PARSER.parse_args()
    PARAMS = yaml.load(ARGS.parameters_file, Loader=yaml.FullLoader)

    SCORECARD = Scorecard(PARAMS)
    SCORECARD.calculate_scorecard_data()



"""Tests the operation of METcalcpy's tost_paired code."""
import statistics
import metcalcpy.util.correlation as pg

from metcalcpy.util.utils import tost_paired


def test_tost_paired():
    x = [103.4, 59.92, 68.17, 94.54, 69.48, 72.17, 74.37, 84.44, 96.74, 94.26, 48.52, 95.68]
    y = [90.11, 77.71, 77.71, 97.51, 58.21, 101.3, 79.84, 96.06, 89.3, 97.22, 61.62, 85.8]

    corr = pg.corr(x=x, y=y)['r'].tolist()[0]

    result = tost_paired(len(x), statistics.mean(x), statistics.mean(y),
                         statistics.stdev(x), statistics.stdev(y), corr,
                         -0.001, 0.001)
    assert result['dif'] == -4.225
    assert result['t'] == (-1.1243156, -1.1312438)
    assert result['p'] == (0.8575922, 0.1410063)
    assert result['degrees_of_freedom'] == 11
    assert result['ci_tost'] == (-10.9529216, 2.5029216)
    assert result['ci_ttest'] == (-12.4705487, 4.0205487)
    assert result['eqbound'] == (-0.0129776, 0.0129776)
    assert result['xlim'] == (-12.298506, 3.848506)
    assert result['combined_outcome'] == 'no_diff_no_eqv'
    assert result['test_outcome'] == 'non-significant'
    assert result['tost_outcome'] == 'non-significant'


if __name__ == "__main__":
    test_tost_paired()


# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** Naval Research Lab
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
# -*- coding: utf-8 -*-
"""
Program Name: calc_difficulty_index.py

Forecast decision difficulty indices.

Implement a set of decision difficulty indices for forecasts of postive
definite quantities such as wind speed and wave height.
"""

import numpy as np
from metcalcpy.piecewise_linear import PiecewiseLinear as plin

__author__ = 'Bill Campbell (NRL) and Lindsay Blank (NCAR)'
__version__ = '0.1.0'


# Enforce positive definiteness of quantities such as standard deviations
EPS = np.finfo(np.float32).eps
# Only allow 2D fields for now
FIELD_DIM = 2

def _input_check(sigmaij, muij, threshold, fieldijn, sigma_over_mu_ref, under_factor):
    """
    Check for valid input to _difficulty_index.

    Parameters
    ----------
    sigmaij : 2D numpy array
        Positive definite array of standard deviations of a 2D field.
    muij : Float scalar or 2D numpy array
        The mean values corresponding to sigmaij.
    threshold : Float (or int) scalar
        A significant value to be compared with values of the forecast field
        for each ensemble member.
    sigma_over_mu_ref : Scalar
    fieldijn : 3D numpy array
        Values of the forecast field. Third dimension is ensemble member.
    under_factor : Float scalar
        Must be between 0 and 1.

    Returns
    -------
    None.

    """
    assert isinstance(threshold, (int, float, np.int32, np.float32))
    assert np.ndim(sigmaij) == FIELD_DIM
    assert np.all(sigmaij) >= EPS
    fieldshape = np.shape(fieldijn)
    # muij is a scalar or 2D array
    if isinstance(muij, np.ndarray):
        # If muij is an array, it must have the same shape as sigmaij
        assert np.shape(muij) == np.shape(sigmaij)
    assert sigma_over_mu_ref >= EPS
    assert np.shape(sigmaij) == tuple(np.squeeze(fieldshape[0:-1]))
    assert isinstance(under_factor, (int, float,
                                     np.int32, np.float32))
    assert 0.0 <= under_factor <= 1.0


def _difficulty_index(sigmaij, muij, threshold, fieldijn, Aplin, sigma_over_mu_ref=EPS, under_factor=0.5):
    """
    Calculates version 6.1 of forecast difficulty index.
    The threshold terms all penalize equal (or slightly unequal) spread.

    Parameters
    ----------
    sigmaij : 2D numpy array
        Positive definite array of standard deviations of a 2D field.
    muij : Float scalar or 2D numpy array
        The mean values corresponding to sigmaij.
    threshold : Float (or int) scalar
        A significant value to be compared with values of the forecast field
        for each ensemble member.
    fieldijn : 3D numpy array
        Values of the forecast field. Third dimension is ensemble member.
    Aplin: PiecewiseLinear object (envelope function)
        Essentially a piecewise linear localization function based on muij
    sigma_over_mu_ref : Scalar, optional
        Highest value of sigmaij/muij for past 5 days (nominally).
    under_factor: Scalar, optional
        Defaults to 0.5 except should be 0.4 for v4 difficulty index

    Returns
    -------
    dij : 2D numpy array
        Normalized v6.1 difficulty index ([0,1.5]). Larger (> 0.5)
        means more difficult.

    """
    # Check for valid input
    _input_check(sigmaij, muij, threshold, fieldijn, sigma_over_mu_ref, under_factor)

    # Variance term in range 0 to 1
    sigma_over_mu = sigmaij / muij
    sigma_over_mu_max = np.nanmax(sigma_over_mu)
    
    # Force reference value to be greater than current max of sigmaij / muij
    sigma_over_mu_ref = np.nanmax([sigma_over_mu_ref,
                                   sigma_over_mu_max])
    variance_term = sigma_over_mu / sigma_over_mu_ref

    # Depends on under_factor.
    under_threshold_count =\
        np.ma.masked_greater_equal(fieldijn, threshold).count(axis=-1)
    nmembers = np.shape(fieldijn)[-1]
    under_prob = under_threshold_count / nmembers

    # Threshold term in range 0 to 1
    threshold_term = 1.0 - np.abs(1.0 - under_factor - under_prob)

    # Linear taper factor
    taper_term = Aplin.values(muij)

    # Difficulty index is the average of the two terms
    # multiplied by a linear taper factor
    dij = 0.5 * taper_term * (variance_term + threshold_term)
    
    return dij


def forecast_difficulty(sigmaij, muij, threshold, fieldijn,
                        Aplin, sigma_over_mu_ref=EPS):
    """
    Calls private function _difficulty_index, 
    to calculate version (v6.1) of forecast difficulty index.

    Parameters
    ----------
    sigmaij : 2D numpy array
        Positive definite array of standard deviations of a 2D field.
    muij : Float scalar or 2D numpy array
        The mean values corresponding to sigmaij.
    threshold : Float (or int) scalar
        A significant value to be compared with values of the forecast field
        for each ensemble member.
    fieldijn : 3D numpy array
        Values of the forecast field. Third dimension is ensemble member.
    Aplin : PiecewiseLinear object (envelope function), optional
        Essentially a piecewise linear localization function based on muij
    sigma_over_mu_ref : Scalar, optional
        Highest value of sigmaij/muij for past 5 days (nominally).

    Returns
    -------
    dij : 2D numpy array
        Normalized difficulty index ([0,1.5]).
        Larger (> 0.5) means more difficult.

    """
    if Aplin is None:
        #  Default to envelope version 6.1
        xunits="feet"                                                                           
        A6_1_name = "A6_1"                                                          
        A6_1_left = 0.0                                                                         
        A6_1_right = 0.0                                                                        
        A6_1_xlist = [3.0, 9.0, 12.0, 21.0]                                                         
        A6_1_ylist = [0.0, 1.5, 1.5, 0.0]                                                   
        Aplin =\
                plin(A6_1_xlist, A6_1_ylist, xunits=xunits,
                        right=A6_1_right, left=A6_1_left,                                       
                        name=A6_1_name)                                                                   
    dij = _difficulty_index(sigmaij, muij, threshold, fieldijn,
                       Aplin, sigma_over_mu_ref)
    return dij

if __name__ == "__main__":
    pass



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: tost_paired.py
"""

from scipy.stats import t, nct
import sys
import math

from metcalcpy.util.utils import round_half_up, PRECISION

CODE_TO_OUTCOME_TO_MESSAGE = {
    'diff_eqv': 'statistically different from zero and statistically equivalent to zero',
    'diff_no_eqv': 'statistically different from zero and statistically not equivalent to zero',
    'no_diff_eqv': 'statistically not different from zero and statistically equivalent to zero',
    'no_diff_no_eqv': 'statistically not different from zero and statistically not equivalent to zero'
}


def pt(q, df, ncp=0, lower_tail=True):
    """
    Calculates the cumulative of the t-distribution

    Args:
        q - vector of quantiles
        df - degrees of freedom (> 0)
        ncp - array_like shape parameters
        lower_tail - if True (default), probabilities are P[X ≤ x], otherwise, P[X > x].

    Returns:
        the cumulative of the t-distribution
    """

    if ncp == 0:
        result = t.cdf(x=q, df=df, loc=0, scale=1)
    else:
        result = nct.cdf(x=q, df=df, nc=ncp, loc=0, scale=1)
    if lower_tail is False:
        result = 1 - result
    return result


def qt(p, df, ncp=0):
    """
    Calculates the quantile function of the t-distribution

     Args:
        p - array_like quantiles
        df - array_like shape parameters
        ncp - array_like shape parameters

    Returns:
        tquantile function of the t-distribution
    """
    if ncp == 0:
        result = t.ppf(q=p, df=df, loc=0, scale=1)
    else:
        result = nct.ppf(q=p, df=df, nc=ncp, loc=0, scale=1)
    return result


def tost_paired(n, m1, m2, sd1, sd2, r12, low_eqbound_dz, high_eqbound_dz, alpha=None):
    """
    TOST function for a dependent t-test (Cohen's dz). Based on Rscript function TOSTpaired

    Args:
        n: sample size (pairs)
        m1: mean of group 1
        m2: mean of group 2
        sd1: standard deviation of group 1
        sd2: standard deviation of group 2
        r12: correlation of dependent variable between group 1 and group 2
        low_eqbound_dz: lower equivalence bounds (e.g., -0.5) expressed in standardized mean difference (Cohen's dz)
        high_eqbound_dz: upper equivalence bounds (e.g., 0.5) expressed in standardized mean difference (Cohen's dz)
        alpha: alpha level (default = 0.05)

    Returns:
        Returns a dictionary with calculated TOST values
                dif - Mean Difference
                t - TOST t-values 1 and 2 as a tuple
                p - TOST p-values and 2 as a tuple
                degrees_of_freedom - degrees of freedom
                ci_tost - confidence interval TOST Lower and Upper limit as a tuple
                ci_ttest - confidence interval TTEST Lower and Upper limit as a tuple
                eqbound - equivalence bound low and high as a tuple
                xlim - limits for x-axis
                combined_outcome - outcome
                test_outcome - pt test outcome
                tist_outcome - TOST outcome

    """
    if not alpha:
        alpha = 0.05
    if low_eqbound_dz >= high_eqbound_dz:
        print(
            'WARNING: The lower bound is equal to or larger than the upper bound.'
            ' Check the plot and output to see if the bounds are specified as you intended.')

    if n < 2:
        print("The sample size should be larger than 1.")
        sys.exit()

    if 1 <= alpha or alpha <= 0:
        print("The alpha level should be a positive value between 0 and 1.")
        sys.exit()
    if sd1 <= 0 or sd2 <= 0:
        print("The standard deviation should be a positive value.")
        sys.exit()
    if 1 < r12 or r12 < -1:
        print("The correlation should be a value between -1 and 1.")
        sys.exit()

    sdif = math.sqrt(sd1 * sd1 + sd2 * sd2 - 2 * r12 * sd1 * sd2)
    low_eqbound = low_eqbound_dz * sdif
    high_eqbound = high_eqbound_dz * sdif
    se = sdif / math.sqrt(n)
    t = (m1 - m2) / se
    degree_f = n - 1

    pttest = 2 * pt(abs(t), degree_f, lower_tail=False)

    t1 = ((m1 - m2) - (low_eqbound_dz * sdif)) / se
    p1 = pt(t1, degree_f, lower_tail=False)
    t2 = ((m1 - m2) - (high_eqbound_dz * sdif)) / se
    p2 = pt(t2, degree_f, lower_tail=True)

    ll90 = ((m1 - m2) - qt(1 - alpha, degree_f) * se)
    ul90 = ((m1 - m2) + qt(1 - alpha, degree_f) * se)
    ptost = max(p1, p2)

    dif = (m1 - m2)
    ll95 = ((m1 - m2) - qt(1 - (alpha / 2), degree_f) * se)
    ul95 = ((m1 - m2) + qt(1 - (alpha / 2), degree_f) * se)
    xlim_l = min(ll90, low_eqbound) - max(ul90 - ll90, high_eqbound - low_eqbound) / 10
    xlim_u = max(ul90, high_eqbound) + max(ul90 - ll90, high_eqbound - low_eqbound) / 10

    if pttest <= alpha and ptost <= alpha:
        combined_outcome = 'diff_eqv'

    if pttest < alpha and ptost > alpha:
        combined_outcome = 'diff_no_eqv'

    if pttest > alpha and ptost <= alpha:
        combined_outcome = 'no_diff_eqv'

    if pttest > alpha and ptost > alpha:
        combined_outcome = 'no_diff_no_eqv'

    if pttest < alpha:
        test_outcome = 'significant'
    else:
        test_outcome = 'non-significant'

    if ptost < alpha:
        tost_outcome = 'significant'
    else:
        tost_outcome = 'non-significant'

    return {
        'dif': round_half_up(dif, PRECISION),
        't': (round_half_up(t1, PRECISION), round_half_up(t2, PRECISION)),
        'p': (round_half_up(p1, PRECISION), round_half_up(p2, PRECISION)),
        'degrees_of_freedom': round_half_up(degree_f, PRECISION),
        'ci_tost': (round_half_up(ll90, PRECISION), round_half_up(ul90, PRECISION) ),
        'ci_ttest': (round_half_up(ll95, PRECISION), round_half_up(ul95, PRECISION)),
        'eqbound': (round_half_up(low_eqbound, PRECISION), round_half_up(high_eqbound, PRECISION)),
        'xlim': (round_half_up(xlim_l, PRECISION), round_half_up(xlim_u, PRECISION)),
        'combined_outcome': combined_outcome,
        'test_outcome': test_outcome,
        'tost_outcome': tost_outcome
    }








# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Contains functions to compute space-time power, coherence and phase spectra.
Mirrors the capabilities of NCL.

List of functions:

mjo_cross_segment:

mjo_cross_segment_realfft:

get_symmasymm:

mjo_cross_coh2pha:

smooth121:

smooth121_1D:

window_cosbell:

mjo_cross:
  The main function to compute cross-spectra is called as
  result = mjo_cross(X,Y,nperseg,segoverlap). This splits the data in X and Y
  into segments of length nperseg, the segments overlap by segoverlap.

kf_filter_mask:

kf_filter:

"""

import numpy as np
from scipy import signal

pi = np.pi
re = 6.371008e6  # Earth's radius in meters
g = 9.80665  # Gravitational acceleration [m s^{-2}]
omega = 7.292e-05  # Angular speed of rotation of Earth [rad s^{-1}]
beta = 2. * omega / re  # beta parameter at the equator


def mjo_cross_segment(XX, YY, opt=False):
    """
    Compute the FFT to get the power and cross-spectra for one time segment.
    :param XX: Input array (time, lat, lon)
    :param YY: Input array (time, lat, lon)
    :param opt: Optional parameter, not currently used. Set to False.
    :return STC: Spectra array of shape (8, nfreq, nwave). Last 4 entries are blank and need to be computed by calling
    mjo_cross_coh2pha. The first 4 entries contain power spectra for XX, power spectra for YY, co-spectra between XX
    and YY, quadrature spectra between XX and YY.
    """
    NT, NM, NL = XX.shape
    # compute fourier decomposition in time and longitude
    Xfft = np.fft.fft2(XX, axes=(0, 2))
    Yfft = np.fft.fft2(YY, axes=(0, 2))
    # normalize by # time samples
    Xfft = Xfft / (NT * NL)
    Yfft = Yfft / (NT * NL)
    # shift 0 frequency and 0 wavenumber to the center
    Xfft = np.fft.fftshift(Xfft, axes=(0, 2))
    Yfft = np.fft.fftshift(Yfft, axes=(0, 2))

    # average the power spectra across all latitudes
    PX = np.average(np.square(np.abs(Xfft)), axis=1)
    PY = np.average(np.square(np.abs(Yfft)), axis=1)

    # compute co- and quadrature spectrum
    PXY = np.average(np.conj(Yfft) * Xfft, axis=1)
    CXY = np.real(PXY)
    QXY = np.imag(PXY)

    PX = PX[:, ::-1]
    PY = PY[:, ::-1]
    CXY = CXY[:, ::-1]
    QXY = QXY[:, ::-1]

    # test if time and longitude are odd or even, fft algorithm
    # returns the Nyquist frequency once for even NT or NL and twice
    # if they are odd
    if NT % 2 == 1:
        nfreq = NT
        if NL % 2 == 1:
            nwave = NL
            STC = np.zeros([8, nfreq, nwave], dtype='double')
            STC[0, :NT, :NL] = PX
            STC[1, :NT, :NL] = PY
            STC[2, :NT, :NL] = CXY
            STC[3, :NT, :NL] = QXY
        else:
            nwave = NL + 1
            STC = np.zeros([8, nfreq, nwave], dtype='double')
            STC[0, :NT, 1:NL + 1] = PX
            STC[1, :NT, 1:NL + 1] = PY
            STC[2, :NT, 1:NL + 1] = CXY
            STC[3, :NT, 1:NL + 1] = QXY
            STC[:, :, 0] = STC[:, :, NL]
    else:
        nfreq = NT + 1
        if NL % 2 == 1:
            nwave = NL
            STC = np.zeros([8, nfreq, nwave], dtype='double')
            STC[0, :NT, :NL] = PX
            STC[1, :NT, :NL] = PY
            STC[2, :NT, :NL] = CXY
            STC[3, :NT, :NL] = QXY
            STC[:, NT, :] = STC[:, 0, :]
        else:
            nwave = NL + 1
            STC = np.zeros([8, nfreq, nwave], dtype='double')
            STC[0, :NT, 1:NL + 1] = PX
            STC[1, :NT, 1:NL + 1] = PY
            STC[2, :NT, 1:NL + 1] = CXY
            STC[3, :NT, 1:NL + 1] = QXY
            STC[0, NT, :] = STC[0, 0, :]
            STC[:, :, 0] = STC[:, :, NL]

    return STC


def mjo_cross_segment_realfft(XX, YY, opt=False):
    """
    Compute the FFT to get the power and cross-spectra for one time segment.
    :param XX: Input array (time, lat, lon)
    :param YY: Input array (time, lat, lon)
    :param opt: Optional parameter, not currently used. Set to False.
    :return STC: Spectra array of shape (8, nfreq, nwave). Last 4 entries are blank and need to be computed by calling
    mjo_cross_coh2pha. The first 4 entries contain power spectra for XX, power spectra for YY, co-spectra between XX
    and YY, quadrature spectra between XX and YY.
    """
    NT, NM, NL = XX.shape

    XX = np.transpose(XX, axes=[1, 2, 0])  # is now (lat, lon, time)
    YY = np.transpose(YY, axes=[1, 2, 0])  # is now (lat, lon, time)

    # compute fourier decomposition in time and longitude
    Xfft = np.fft.rfft2(XX, axes=(1, 2))  # (lat, nlon, ntim)
    Yfft = np.fft.rfft2(YY, axes=(1, 2))   # (lat, nlon, ntim)

    # return array to (time, lat, lon)
    Xfft = np.transpose(Xfft, axes=[2, 0, 1])
    Yfft = np.transpose(Yfft, axes=[2, 0, 1])

    # normalize by # time samples
    Xfft = Xfft / (NT * NL)
    Yfft = Yfft / (NT * NL)

    # shift 0 wavenumber to the center
    Xfft = np.fft.fftshift(Xfft, axes=2)
    Yfft = np.fft.fftshift(Yfft, axes=2)

    # average the power spectra across all latitudes
    PX = np.average(np.square(np.abs(Xfft)), axis=1)
    PY = np.average(np.square(np.abs(Yfft)), axis=1)

    # compute co- and quadrature spectrum
    PXY = np.conj(Xfft) * Yfft
    CXY = np.average(np.real(PXY), axis=1)
    QXY = np.average(np.imag(PXY), axis=1)

    PX = PX[:, ::-1]
    PY = PY[:, ::-1]
    CXY = CXY[:, ::-1]
    QXY = QXY[:, ::-1]

    # test if time and longitude are odd or even, fft algorithm
    # returns the Nyquist frequency once for even NT or NL and twice
    # if they are odd
    NT = int(NT / 2) + 1
    if NT % 2 == 1:
        nfreq = NT
        if NL % 2 == 1:
            nwave = NL
            STC = np.zeros([8, nfreq, nwave], dtype='double')
            STC[0, :NT, :NL] = PX
            STC[1, :NT, :NL] = PY
            STC[2, :NT, :NL] = CXY
            STC[3, :NT, :NL] = QXY
        else:
            nwave = NL + 1
            STC = np.zeros([8, nfreq, nwave], dtype='double')
            STC[0, :NT, 1:NL + 1] = PX
            STC[1, :NT, 1:NL + 1] = PY
            STC[2, :NT, 1:NL + 1] = CXY
            STC[3, :NT, 1:NL + 1] = QXY
            STC[:, :, 0] = STC[:, :, NL]
    else:
        nfreq = NT + 1
        if NL % 2 == 1:
            nwave = NL
            STC = np.zeros([8, nfreq, nwave], dtype='double')
            STC[0, :NT, :NL] = PX
            STC[1, :NT, :NL] = PY
            STC[2, :NT, :NL] = CXY
            STC[3, :NT, :NL] = QXY
            STC[:, NT, :] = STC[:, 0, :]
        else:
            nwave = NL + 1
            STC = np.zeros([8, nfreq, nwave], dtype='double')
            STC[0, :NT, 1:NL + 1] = PX
            STC[1, :NT, 1:NL + 1] = PY
            STC[2, :NT, 1:NL + 1] = CXY
            STC[3, :NT, 1:NL + 1] = QXY
            STC[:, NT, :] = STC[:, 0, :]
            STC[:, :, 0] = STC[:, :, NL]

    return STC


def get_symmasymm(X, lat, opt=False):
    """
    Split the data in X into symmetric and anti-symmetric
    (across the equator) parts. Return only the part we are
    interested in.
    :param X: Array (time, lat, lon).
    :param lat: Latitude values corresponding to dimension 1 of X.
    :param opt: Parameter to choose symmetric or anti-symmetric part across the equator.
    :return : symmetric or anti-symmetric part of X
    """
    if opt:
        NT, NM, NL = X.shape
        if opt == 'symm':
            x = X[:, lat[:] >= 0, :]
            if len(lat) % 2 == 1:
                for ll in range(NM // 2 + 1):
                    x[:, ll, :] = 0.5 * (X[:, ll, :] + X[:, NM - ll - 1, :])
            else:
                for ll in range(NM // 2):
                    x[:, ll, :] = 0.5 * (X[:, ll, :] + X[:, NM - ll - 1, :])
        else:
            if opt == 'asymm' or opt == 'anti-symm':
                x = X[:, lat[:] > 0, :]
                if len(lat) % 2 == 1:
                    for ll in range(NM // 2):
                        x[:, ll, :] = 0.5 * (X[:, ll, :] - X[:, NM - ll - 1, :])
                else:
                    for ll in range(NM // 2):
                        x[:, ll, :] = 0.5 * (X[:, ll, :] - X[:, NM - ll - 1, :])
    else:
        print("Please provide a valid option: symm or asymm.")

    return x


def mjo_cross_coh2pha(STC, opt=False):
    """
    Compute coherence squared and phase spectrum from averaged power and
    cross-spectral estimates.
    :param STC: Spectra array.
    :return STC: Spectra array of the same size with entries 4-7 (coherence squared, phase angle, phase component 1,
    phase component 2) recomputed based on the power and cross-spectra in entries 0-3.
    """

    nvar, nfreq, nwave = STC.shape

    PX = STC[0, :, :]
    PY = STC[1, :, :]
    CXY = STC[2, :, :]
    QXY = STC[3, :, :]

    PY[PY == 0] = np.nan
    COH2 = (np.square(CXY) + np.square(QXY)) / (PX * PY)
    PHAS = np.arctan2(QXY, CXY)

    V1 = -QXY / np.sqrt(np.square(QXY) + np.square(CXY))
    V2 = CXY / np.sqrt(np.square(QXY) + np.square(CXY))

    STC[4, :, :] = COH2
    STC[5, :, :] = PHAS
    STC[6, :, :] = V1
    STC[7, :, :] = V2

    return (STC)


def smooth121(STC, freq, opt=False):
    """
    This function takes a coherence spectra array STC [ 8, nfreq, nwave] and smoothes the first
    4 entries ( Xpower, Ypower, Co-spectrum, Quadrature-spectrum ). Entries 4-7 of STC need to
    be recomputed after smoothing to have the matching phase, coherence-squared, and phase angle
    components.
    Smooth only in frequency and only positive frequencies.
    :param array_in:
        Input array
    :type array_in: np array
    :return: STC
    :rtype: np array
    """
    nvar, nfreq, nwave = STC.shape

    # find time-mean index
    indfreqzero = int(np.where(freq == 0.)[0])

    for wv in range(0, nwave):
        STC[0, indfreqzero + 1:, wv] = smooth121_1D(STC[0, indfreqzero + 1:, wv])
        STC[1, indfreqzero + 1:, wv] = smooth121_1D(STC[1, indfreqzero + 1:, wv])
        STC[2, indfreqzero + 1:, wv] = smooth121_1D(STC[2, indfreqzero + 1:, wv])
        STC[3, indfreqzero + 1:, wv] = smooth121_1D(STC[3, indfreqzero + 1:, wv])

    return (STC)


def smooth121_1D(array_in):
    """
    Smoothing function that takes a 1D array and passes it through a 1-2-1 filter.
    This function is a modified version of the wk_smooth121 from NCL.
    The weights for the first and last points are  3-1 (1st) or 1-3 (last) conserving the total sum.
    :param array_in:
        Input array
    :type array_in: np array
    :return: array_out
    :rtype: np array
    """

    temp = np.copy(array_in)
    array_out = np.copy(temp) * 0.0
    #weights = np.array([1.0, 2.0, 1.0]) / 4.0
    #sma = np.convolve(temp, weights, 'valid')
    #array_out[1:-1] = sma

    for i in np.arange(0, len(temp), 1):
        if np.isnan(temp[i]):
            array_out[0] = np.nan
        elif i == 0 or np.isnan(temp[i-1]):
            array_out[i] = (3*temp[i]+temp[i+1])/4
        elif i == (len(temp)-1) or np.isnan(temp[i+1]):
            array_out[i] = (3 * temp[i] + temp[i-1]) / 4
        else:
            array_out[i] = (temp[i+1] + 2 * temp[i] + temp[i-1]) / 4

    return array_out


def window_cosbell(N, pct, opt=False):
    """
    Compute an equivalent tapering window to the NCL taper function.
    :param N: Length of the time series to be tapered.
    :param pct: Percent of the time series to taper.
    :return x: Array of length N and values 1 with pct/2 beginning and end values tapered to zero.
    """

    x = np.ones(N, dtype='double')
    M = int((pct * N + 0.5) / 2)
    if M < 1:
        M = 1

    for i in range(1, M + 1):
        wgt = 0.5 - 0.5 * np.cos(np.pi / M * (i - 0.5))
        x[i - 1] = x[i - 1] * wgt
        x[-i] = x[-i] * wgt

    return x


def mjo_cross(X, Y, segLen, segOverLap, opt=False):
    """
    MJO cross spectrum function. This function calls the above functions to compute
    cross spectral estimates for each segment of length segLen. Segments overlap by
    segOverLap. This function mirrors the NCL routine mjo_cross.
    Return value is a dictionary.
    :param X: Input data 3D array ( time, lat, lon).
    :param Y: Input data 3D array ( time, lat, lon).
    :param segLen: Length of the time segments.
    :param segOverLap: Length of the overlap between time segments.
    :param opt: Optional parameter. Not currently used, set to False.
    :return dict: Dictionary containing the spectral array (STC), frequency array (freq), zonal wavenumber array (wave),
    the number of segments used (nseg), the estimated degrees of freedom (dof), the probability levels (p),
    the coherence squared values corresponding to the probability levels (prob_coh2)
    """

    ntim, nlat, mlon = X.shape
    ntim1, nlat1, mlon1 = Y.shape

    if any([ntim - ntim1, nlat - nlat1, mlon - mlon1]) != 0:
        print("mjo_cross: X and Y must be same size")
        print("           dimX=" + [ntim, nlat, mlon] + "   dimY=" + [ntim1, nlat1, mlon1])
        return

    # make a local copy (time,lat,lon)
    x = X
    y = Y

    # detrend overall series in time, dimension 0
    x = signal.detrend(x, 0)
    y = signal.detrend(y, 0)

    # generate Tukey window, taper 10% of series
    pct = 0.10
    window = window_cosbell(segLen, pct)
    window = np.tile(window, [1, 1, 1])
    window = np.transpose(window)
    window = np.tile(window, [1, nlat, mlon])

    # test if time and longitude are odd or even, fft algorithm
    # returns the Nyquist frequency once for even NT or NL and twice
    # if they are odd
    if (segLen/2+1) % 2 == 1:
        nfreq = int(segLen/2) + 1
    else:
        nfreq = int(segLen/2) + 2
    if mlon % 2 == 1:
        nwave = mlon
    else:
        nwave = mlon + 1

    # initialize spectrum array
    STC = np.zeros([8, nfreq, nwave], dtype='double')
    wave = np.arange(-int(nwave / 2), int(nwave / 2) + 1, 1.)
    freq = np.linspace(0, 0.5, num=nfreq)

    # find time-mean index
    indfreq0 = np.where(freq == 0.)[0]

    # loop through segments and compute cross-spectra
    kseg = 0
    ntStrt = 0
    switch = True
    while switch:
        ntLast = ntStrt + segLen
        if (ntLast > (ntim - 1)):
            switch = False
            continue

        XX = x[ntStrt:ntLast, :, :] * window
        YY = y[ntStrt:ntLast, :, :] * window
        STCseg = mjo_cross_segment_realfft(XX, YY, 0)
        # set time-mean power to NaN
        STCseg[:, indfreq0, :] = np.nan
        # apply 1-2-1 smoother in frequency
        smooth121(STCseg, freq)
        # sum segment spectra
        STC = STC + STCseg

        kseg = kseg + 1
        ntStrt = ntLast + segOverLap - 1

    STC = STC / kseg

    # compute phase and coherence from averaged spectra
    mjo_cross_coh2pha(STC)

    # conservative estimate for DOFs, 2.667 is for 1-2-1 smoother
    dof = 2.667 * kseg
    p = [0.80, 0.85, 0.90, 0.925, 0.95, 0.99]  # probability levels
    prob = p
    prob_coh2 = 1 - (1 - np.power(p, (0.5 * dof - 1)))

    return {'STC': STC, 'freq': freq, 'wave': wave, 'nseg': kseg, 'dof': dof, 'p': prob, 'prob_coh2': prob_coh2}


def kf_filter_mask(fftIn, obsPerDay, tMin, tMax, kMin, kMax, hMin, hMax, waveName):
    """
    Generate a filtered mask array based on the FFT array and the wave information. Set all values
    outside the specified wave dispersion curves to zero.
    :param fftData: Array of fft coefficients ( wavenumber x freq ), has to be 2 dimensional.
    :param obsPerDay: Number of observations per day.
    :param tMin: Minimum period to include in filtering region.
    :param tMax: Maximum period to include in filtering region.
    :param kMin: Minimum wavenumber to include in filtering region.
    :param kMax: Maximum wavenumber to include in filtering region.
    :param hMin: Minimum equivalent depth to include in filtering region.
    :param hMax: Maximum equivalent depth to include in filtering region.
    :param waveName: Name of the wave to filter for.
    :return: Array containing the fft coefficients of the same size as the input data, with coefficients outside the
    desired region set to zero.
    """
    fftData = np.copy(fftIn)
    fftData = np.transpose(fftData)
    nf, nk = fftData.shape  # frequency, wavenumber array
    fftData = fftData[:, ::-1]

    nt = (nf - 1) * 2
    jMin = int(round(nt / (tMax * obsPerDay)))
    jMax = int(round(nt / (tMin * obsPerDay)))
    jMax = np.array([jMax, nf]).min()

    if kMin < 0:
        iMin = int(round(nk + kMin))
        iMin = np.array([iMin, nk // 2]).max()
    else:
        iMin = int(round(kMin))
        iMin = np.array([iMin, nk // 2]).min()

    if kMax < 0:
        iMax = int(round(nk + kMax))
        iMax = np.array([iMax, nk // 2]).max()
    else:
        iMax = int(round(kMax))
        iMax = np.array([iMax, nk // 2]).min()

    # set the appropriate coefficients outside the frequency range to zero
    # print(fftData[:, 0])
    if jMin > 0:
        fftData[0:jMin, :] = 0
    if jMax < nf:
        fftData[jMax + 1:nf, :] = 0
    if iMin < iMax:
        # Set things outside the wavenumber range to zero, this is more normal
        if iMin > 0:
            fftData[:, 0:iMin] = 0
        if iMax < nk:
            fftData[:, iMax + 1:nk] = 0
    else:
        # Set things inside the wavenumber range to zero, this should be somewhat unusual
        fftData[:, iMax + 1:iMin] = 0

    c = np.empty([2])
    if hMin == -9999:
        c[0] = np.nan
        if hMax == -9999:
            c[1] = np.nan
    else:
        if hMax == -9999:
            c[1] = np.nan
        else:
            c = np.sqrt(g * np.array([hMin, hMax]))

    spc = 24 * 3600. / (2 * pi * obsPerDay)  # seconds per cycle

    # Now set things to zero that are outside the wave dispersion. Loop through wavenumbers
    # and find the limits for each one.
    for i in range(nk):
        if i < (nk / 2):
            # k is positive
            k = i / re
        else:
            # k is negative
            k = -(nk - i) / re

        freq = np.array([0, nf]) / spc
        jMinWave = 0
        jMaxWave = nf
        if (waveName == "Kelvin") or (waveName == "kelvin") or (waveName == "KELVIN"):
            ftmp = k * c
            freq = np.array(ftmp)
        if (waveName == "ER") or (waveName == "er"):
            ftmp = -beta * k / (k ** 2 + 3 * beta / c)
            freq = np.array(ftmp)
        if (waveName == "MRG") or (waveName == "IG0") or (waveName == "mrg") or (waveName == "ig0"):
            if k == 0:
                ftmp = np.sqrt(beta * c)
                freq = np.array(ftmp)
            else:
                if k > 0:
                    ftmp = k * c * (0.5 + 0.5 * np.sqrt(1 + 4 * beta / (k ** 2 * c)))
                    freq = np.array(ftmp)
                else:
                    ftmp = k * c * (0.5 - 0.5 * np.sqrt(1 + 4 * beta / (k ** 2 * c)))
                    freq = np.array(ftmp)
        if (waveName == "IG1") or (waveName == "ig1"):
            ftmp = np.sqrt(3 * beta * c + k ** 2 * c ** 2)
            freq = np.array(ftmp)
        if (waveName == "IG2") or (waveName == "ig2"):
            ftmp = np.sqrt(5 * beta * c + k ** 2 * c ** 2)
            freq = np.array(ftmp)

        if hMin == -9999:
            jMinWave = 0
        else:
            jMinWave = int(np.floor(freq[0] * spc * nt))
        if hMax == -9999:
            jMaxWave = nf
        else:
            jMaxWave = int(np.ceil(freq[1] * spc * nt))
        jMaxWave = np.array([jMaxWave, 0]).max()
        jMinWave = np.array([jMinWave, nf]).min()

        # set appropriate coefficients to zero
        if jMinWave > 0:
            fftData[0:jMinWave, i] = 0
        if jMaxWave < nf:
            fftData[jMaxWave + 1:nf, i] = 0

    fftData = fftData[:, ::-1]
    fftData = np.transpose(fftData)

    return fftData


def kf_filter(data, obsPerDay, tMin, tMax, kMin, kMax, hMin, hMax, waveName):
    """
    Filter 2D (time x lon) input data for a convectively coupled equatorial wave region in
    wavenumber - frequency space.
    :param Data: Input data ( time x lon ), has to be 2 dimensional.
    :param obsPerDay: Number of observations per day.
    :param tMin: Minimum period to include in filtering region.
    :param tMax: Maximum period to include in filtering region.
    :param kMin: Minimum wavenumber to include in filtering region.
    :param kMax: Maximum wavenumber to include in filtering region.
    :param hMin: Minimum equivalent depth to include in filtering region.
    :param hMax: Maximum equivalent depth to include in filtering region.
    :param waveName: Name of the wave to filter for.
    :return: Array containing the filtered data of the same size as the input data.
    """

    # reorder to (lon x time) to be able to use rfft on the time dimension
    data = np.transpose(data, axes=[1, 0])
    fftdata = np.fft.rfft2(data, axes=(0, 1))

    fftfilt = kf_filter_mask(fftdata, obsPerDay, tMin, tMax, kMin, kMax, hMin, hMax, waveName)

    datafilt = np.fft.irfft2(fftfilt, axes=(0, 1))
    datafilt = np.transpose(datafilt, axes=[1, 0])

    return datafilt



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: nbrctc_statistics.py
"""

from metcalcpy.util.ctc_statistics import calculate_baser, calculate_acc, calculate_fbias, \
    calculate_fmean, calculate_pody, calculate_pofd, calculate_podn, calculate_far, calculate_csi, \
    calculate_gss, calculate_hk, calculate_hss, calculate_odds
from metcalcpy.util.utils import round_half_up, sum_column_data_by_name, PRECISION

__author__ = 'Tatiana Burek'
__version__ = '0.1.0'


def calculate_nbr_baser(input_data, columns_names):
    """Performs calculation of NBR_BASER - Base rate

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated NBR_BASER as float
            or None if some of the data values are missing or invalid
    """

    return calculate_baser(input_data, columns_names)


def calculate_nbr_acc(input_data, columns_names):
    """Performs calculation of NBR_ACC - Accuracy

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated NBR_ACC as float
            or None if some of the data values are missing or invalid
    """

    return calculate_acc(input_data, columns_names)


def calculate_nbr_fbias(input_data, columns_names):
    """Performs calculation of NBR_FBIAS - Frequency Bias

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated NBR_FBIAS as float
            or None if some of the data values are missing or invalid
    """

    return calculate_fbias(input_data, columns_names)


def calculate_nbr_fmean(input_data, columns_names):
    """Performs calculation of NBR_FMEAN - Forecast mean

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated NBR_FMEAN as float
            or None if some of the data values are missing or invalid
    """

    return calculate_fmean(input_data, columns_names)


def calculate_nbr_pody(input_data, columns_names):
    """Performs calculation of NBR_PODY - Probability of detecting yes

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated NBR_PODY as float
            or None if some of the data values are missing or invalid
    """

    return calculate_pody(input_data, columns_names)


def calculate_nbr_pofd(input_data, columns_names):
    """Performs calculation of NBR_POFD - Probability of false detection

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated NBR_POFD as float
            or None if some of the data values are missing or invalid
    """

    return calculate_pofd(input_data, columns_names)


def calculate_nbr_podn(input_data, columns_names):
    """Performs calculation of NBR_PODN - Probability of false detection

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated NBR_PODN as float
            or None if some of the data values are missing or invalid
    """

    return calculate_podn(input_data, columns_names)


def calculate_nbr_far(input_data, columns_names):
    """Performs calculation of NBR_FAR - False alarm ratio

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated NBR_FAR as float
            or None if some of the data values are missing or invalid
    """

    return calculate_far(input_data, columns_names)


def calculate_nbr_csi(input_data, columns_names):
    """Performs calculation of NBR_CSI - Critical Success Index

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated NBR_CSI as float
            or None if some of the data values are missing or invalid
    """

    return calculate_csi(input_data, columns_names)


def calculate_nbr_gss(input_data, columns_names):
    """Performs calculation of NBR_GSS - Gilbert Skill Score

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated NBR_GSS as float
            or None if some of the data values are missing or invalid
    """

    return calculate_gss(input_data, columns_names)


def calculate_nbr_hk(input_data, columns_names):
    """Performs calculation of NBR_HK - Hanssen-Kuipers Discriminant

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated NBR_HK as float
            or None if some of the data values are missing or invalid
    """

    return calculate_hk(input_data, columns_names)


def calculate_nbr_hss(input_data, columns_names):
    """Performs calculation of NBR_HSS - Heidke Skil lScore

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated NBR_HSS as float
            or None if some of the data values are missing or invalid
    """

    return calculate_hss(input_data, columns_names)


def calculate_nbr_odds(input_data, columns_names):
    """Performs calculation of NBR_ODDS - Odds Ratio

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated NBR_ODDS as float
            or None if some of the data values are missing or invalid
    """

    return calculate_odds(input_data, columns_names)


def calculate_nbr_ctc_total(input_data, columns_names):
    """Performs calculation of Total number of matched pairs for
        Neighborhood Contingency Table Statistics
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated Total number of matched pairs as float
            or None if some of the data values are missing or invalid
    """
    total = sum_column_data_by_name(input_data, columns_names, 'total')
    return round_half_up(total, PRECISION)



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
# -*- coding: utf-8 -*-

"""
Utilities for plotting the Matsuno Dispersion Curves
Created by: Alejandro Jaramillo
ajaramillomoreno@gmail.com
2018


References:

Matsuno, T. (1966). Quasi-Geostrophic Motions in the Equatorial Area.
Journal of the Meteorological Society of Japan.
Ser. II, 44(1), 25–43.
https://doi.org/10.2151/jmsj1965.44.1_25

Wheeler, M., & Kiladis, G. N. (1999).
Convectively Coupled Equatorial Waves: Analysis of Clouds and Temperature in
the Wavenumber–Frequency Domain. Journal of the Atmospheric Sciences,
56(3), 374–399.
https://doi.org/10.1175/1520-0469(1999)056<0374:CCEWAO>2.0.CO;2

Kiladis, G. N., Wheeler, M. C., Haertel, P. T., Straub, K. H.,
& Roundy, P. E. (2009). Convectively coupled equatorial waves.
Reviews of Geophysics, 47(2), RG2003.
https://doi.org/10.1029/2008RG000266

Wheeler, M. C., & Nguyen, H. (2015).
TROPICAL METEOROLOGY AND CLIMATE | Equatorial Waves.
In Encyclopedia of Atmospheric Sciences (pp. 102–112). Elsevier.
https://doi.org/10.1016/B978-0-12-382225-3.00414-X
"""
import numpy as np
from scipy.optimize import fsolve
from functools import reduce
import pandas as pd
import matplotlib.pyplot as plt
pi = np.pi
re    = 6.371008e6 # Earth's radius in meters
g     = 9.80665 # Gravitational acceleration [m s^{-2}]
omega = 7.292e-05 # Angular speed of rotation of Earth [rad s^{-1}]
deg2rad = pi/180 # Degrees to Radians
sec2day = 1./(24.*60.*60.) #Seconds to Days

def beta_parameters(latitude):
    """
    Function that calculates the beta-plane parameters that are a Functions
    of the latitude, where **beta** is the beta-plane parameter, **perimeter**
    is the perimeter in meters of the Earths's circunference at the given
    latitude.
    :param latitude:
        Latitude
    :type latitude: Float
    :return: (beta, perimeter)
    :rtype: tuple
    """
    beta  = 2.*omega*np.cos(abs(latitude)*deg2rad)/re
    perimeter = 2.*pi*re*np.cos(abs(latitude)*deg2rad)
    return (beta,perimeter)

def wn_array(max_wn,n_wn):
    """
    Creates an array with wavenumbers in the range (-max_wn,max_wn).
    :param max_wn:
        Max global wave number.
        The global wave number range is (-max_wn,max_wn)
    :param n_wn:
        Number of global wave numbers in the range (-max_wn,max_wn)
    :type maxwn: Positive Integer (max_wn > 0)
    :type n_wn: Integer
    :return: Array of Global Wavenumbers
    :rtype: Numpy Array
    """
    maxwn = abs(int(max_wn)) # maxwn is Positive Integer (maxwn > 0)
    n_wn = abs(int(n_wn))
    wn = np.linspace(-max_wn,max_wn,n_wn)
    return wn

def wn2k(wn,perimeter):
    """
    Converts an array of Global wave numbers to wavenumber in [rad m^{-1}].
    :param wn:
        Array of Global wavenumbers.
    :param perimeter:
        Perimeter in meters of the Earths's circunference at the given latitude
    :type wn: Numpy Array
    :type perimeter: Float
    :return: Array of Wavenumbers in [rad m^{-1}]
    :rtype: Numpy Array
    """
    wavelength = perimeter/wn # Wavekength[m]
    k  = 2.*pi/wavelength # Wavenumber[rad m^{-1}]
    return k

def afreq2freq(angular_frequency):
    """
    Convert angular frequency in [rad s^{-1}] to frequency in Cycles per
    Day(CPD).
    :param angular_frequency:
        Angular Frequency
    :type angular_frequency: Numpy Array
    :return: (Period,Frequency)
        Period in [days/cycle]
        Frequency in [cycles/day] Cycles per Day(CPD)
    :rtype: tuple
    """
    period = (2.*pi/angular_frequency)*sec2day # Period in [days/cycle]
    frequency = 1./period #[cycles/day] Cycles per Day(CPD)
    return (period,frequency)

def kelvin_mode(he,latitude=0,max_wn=50,n_wn=500):
    """
    Function that calculates the dispersion curve for the Equatorial Kelvin
    Wave for a given Equivalent Depth.
    :param he:
        Equivalent Depth
    :param latitude:
        Latitude
    :param max_wn:
        Max global wave number.
        The global wave number range is (-max_wn,max_wn)
    :param n_wn:
        Number of global wave numbers in the range (-max_wn,max_wn)
    :type he: Float
    :type latitude: Float
    :type maxwn: Positive Integer (max_wn > 0)
    :type n_wn: Integer
    :return: DataFrame with wn and frequency
    :rtype: DataFrame
    """
    (beta,perimeter) = beta_parameters(latitude)
    wn = wn_array(max_wn,n_wn) #Global Wavenumber
    k  = wn2k(wn,perimeter) # Wavenumber[rad m^{-1}]
    k[k<=0] = np.nan
    angular_frequency = np.sqrt(g*he)*k # [rad s^{-1}]
    (period,frequency) = afreq2freq(angular_frequency)
    # Period in [days/cycle]
    # Frequency [cycles/day] Cycles per Day(CPD)
    name = 'Kelvin(he='+str(he)+'m)'
    df = pd.DataFrame(data={name:frequency},index=wn)
    df.index.name = 'Wavenumber'
    return df

def mrg_mode(he,latitude=0,max_wn=50,n_wn=500):
    """
    Function that calculates the dispersion curve for the Equatorial Mixed
    Rossby Gravity Wave for a given Equivalent Depth.
    :param he:
        Equivalent Depth
    :param latitude:
        Latitude
    :param max_wn:
        Max global wave number.
        The global wave number range is (-max_wn,max_wn)
    :param n_wn:
        Number of global wave numbers in the range (-max_wn,max_wn)
    :type he: Float
    :type latitude: Float
    :type maxwn: Positive Integer (max_wn > 0)
    :type n_wn: Integer
    :return: DataFrame with wn and frequency
    :rtype: DataFrame
    """
    (beta,perimeter) = beta_parameters(latitude)
    wn = wn_array(max_wn,n_wn) #Global Wavenumber
    #wn = wn[wn<0] # Extract only wn < 0
    k  = wn2k(wn,perimeter) # Wavenumber[rad m^{-1}]
    k[k>=0] = np.nan
    angular_frequency = np.sqrt(g*he)*k*(0.5-0.5*np.sqrt(1.+\
                        (4*beta/(k*k*np.sqrt(g*he))))) # [rad s^{-1}]
    (period,frequency) = afreq2freq(angular_frequency)
    # Period in [days/cycle]
    # Frequency [cycles/day] Cycles per Day(CPD)
    name = 'MRG(he='+str(he)+'m)'
    df = pd.DataFrame(data={name:frequency},index=wn)
    df.index.name = 'Wavenumber'
    return df
def eig_n_0(he,latitude=0,max_wn=50,n_wn=500):
    """
    Function that calculates the dispersion curve for the Equatorial Mixed
    Rossby Gravity Wave for a given Equivalent Depth.
    :param he:
        Equivalent Depth
    :param latitude:
        Latitude
    :param max_wn:
        Max global wave number.
        The global wave number range is (-max_wn,max_wn)
    :param n_wn:
        Number of global wave numbers in the range (-max_wn,max_wn)
    :type he: Float
    :type latitude: Float
    :type maxwn: Positive Integer (max_wn > 0)
    :type n_wn: Integer
    :return: DataFrame with wn and frequency
    :rtype: DataFrame
    """
    (beta,perimeter) = beta_parameters(latitude)
    wn = wn_array(max_wn,n_wn) #Global Wavenumber
    k  = wn2k(wn,perimeter) # Wavenumber[rad m^{-1}]
    k[k<=0] = np.nan
    angular_frequency = np.sqrt(g*he)*k*(0.5+0.5*np.sqrt(1.+\
                        (4*beta/(k*k*np.sqrt(g*he))))) # [rad s^{-1}]
    (period,frequency) = afreq2freq(angular_frequency)
    # Period in [days/cycle]
    # Frequency [cycles/day] Cycles per Day(CPD)
    name = 'EIG(n=0,he='+str(he)+'m)'
    df = pd.DataFrame(data={name:frequency},index=wn)
    df.index.name = 'Wavenumber'
    return df

def er_n(he,n,latitude=0.,max_wn=50,n_wn=500):
    """
    Function that calculates the dispersion curve for the Equatorial Mixed
    Rossby Gravity Wave for a given Equivalent Depth.
    :param he:
        Equivalent Depth
    :param n:
        Meridional Mode Number
    :param latitude:
        Latitude
    :param max_wn:
        Max global wave number.
        The global wave number range is (-max_wn,max_wn)
    :param n_wn:
        Number of global wave numbers in the range (-max_wn,max_wn)
    :type he: Float
    :type n: Integer
    :type latitude: Float
    :type maxwn: Positive Integer (max_wn > 0)
    :type n_wn: Integer
    :return: DataFrame with wn and frequency
    :rtype: DataFrame
    """
    (beta,perimeter) = beta_parameters(latitude)
    wn = wn_array(max_wn,n_wn) #Global Wavenumber
    k  = wn2k(wn,perimeter) # Wavenumber[rad m^{-1}]
    # Use the Approximation to the Equatorial Rossby dispersion relationship as
    # a seed for the solver function
    angular_frequency = -beta*k/((k*k)+(2.*n+1.)*(beta/np.sqrt(g*he)))
    angular_frequency[k>=0] = np.nan
    angular_frequency[k<0] = fsolve(dispersion,angular_frequency[k<0],\
                                    args=(k[k<0],n,he,beta))
    (period,frequency) = afreq2freq(angular_frequency)
    # Period in [days/cycle]
    # Frequency [cycles/day] Cycles per Day(CPD)
    name = 'ER(n='+str(n)+',he='+str(he)+'m)'
    df = pd.DataFrame(data={name:frequency},index=wn)
    df.index.name = 'Wavenumber'
    return df

def eig_n(he,n,latitude=0.,max_wn=50,n_wn=500):
    """
    Function that calculates the dispersion curve for the Equatorial Easterly
    Gravity Wave for a given Equivalent Depth.
    :param he:
        Equivalent Depth
    :param n:
        Meridional Mode Number
    :param latitude:
        Latitude
    :param max_wn:
        Max global wave number.
        The global wave number range is (-max_wn,max_wn)
    :param n_wn:
        Number of global wave numbers in the range (-max_wn,max_wn)
    :type he: Float
    :type n: Integer
    :type latitude: Float
    :type maxwn: Positive Integer (max_wn > 0)
    :type n_wn: Integer
    :return: DataFrame with wn and frequency
    :rtype: DataFrame
    """
    (beta,perimeter) = beta_parameters(latitude)
    wn = wn_array(max_wn,n_wn) #Global Wavenumber
    k  = wn2k(wn,perimeter) # Wavenumber[rad m^{-1}]
    # Use the Approximation to the EIG dispersion relationship as
    # a seed for the solver function
    angular_frequency = np.sqrt((2.*n+1.)*beta*np.sqrt(g*he)+(k**2)*g*he)
    angular_frequency[k<=0] = np.nan
    angular_frequency[k>0] = fsolve(dispersion,angular_frequency[k>0],\
                                    args=(k[k>0],n,he,beta))
    (period,frequency) = afreq2freq(angular_frequency)
    # Period in [days/cycle]
    # Frequency [cycles/day] Cycles per Day(CPD)
    name = 'EIG(n='+str(n)+',he='+str(he)+'m)'
    df = pd.DataFrame(data={name:frequency},index=wn)
    df.index.name = 'Wavenumber'
    return df

def wig_n(he,n,latitude=0.,max_wn=50,n_wn=500):
    """
    Function that calculates the dispersion curve for the Equatorial Westerly
    Gravity Wave for a given Equivalent Depth.
    :param he:
        Equivalent Depth
    :param n:
        Meridional Mode Number
    :param latitude:
        Latitude
    :param max_wn:
        Max global wave number.
        The global wave number range is (-max_wn,max_wn)
    :param n_wn:
        Number of global wave numbers in the range (-max_wn,max_wn)
    :type he: Float
    :type n: Integer
    :type latitude: Float
    :type maxwn: Positive Integer (max_wn > 0)
    :type n_wn: Integer
    :return: DataFrame with wn and frequency
    :rtype: DataFrame
    """
    (beta,perimeter) = beta_parameters(latitude)
    wn = wn_array(max_wn,n_wn) #Global Wavenumber
    k  = wn2k(wn,perimeter) # Wavenumber[rad m^{-1}]
    # Use the Approximation to the WIG dispersion relationship as
    # a seed for the solver function
    angular_frequency = np.sqrt((2.*n+1.)*beta*np.sqrt(g*he)+(k**2)*g*he)
    angular_frequency[k>=0] = np.nan
    angular_frequency[k<0] = fsolve(dispersion,angular_frequency[k<0],\
                                    args=(k[k<0],n,he,beta))
    (period,frequency) = afreq2freq(angular_frequency)
    # Period in [days/cycle]
    # Frequency [cycles/day] Cycles per Day(CPD)
    name = 'WIG(n='+str(n)+',he='+str(he)+'m)'
    df = pd.DataFrame(data={name:frequency},index=wn)
    df.index.name = 'Wavenumber'
    return df

def dispersion(w,k,n,he,beta):
    """
    Dispersion relationship for Matsuno Modes(See Wheeler and Nguyen (2015)
    , Eq (13)).
    The roots of this function correspond to the angular frequencies of the
    Matsuno modes for a given k.
    :param w:
        Angular Frequency
    :param k:
        Longitudinal Wavenumber
    :param n:
        Meridional Mode Number
    :param he:
        Equivalent Depth
    :param beta:
        Beta-Plane Parameter
    :type w: Float
    :type k: Float
    :type n: Integer
    :type he: Float
    :type beta: Float
    :return: Zero if w and k corresponds to a Matsuno Mode.
    :rtype: Float
    """
    disp = w**3-g*he*(k**2+(beta*(2.*n+1.)/np.sqrt(g*he)))*w-k*beta*g*he
    return disp

def matsuno_dataframe(he,n=[1,2,3],latitude=0.,max_wn=50,n_wn=500):
    """
    Creates a dataframe with all Matsuno modes for a given set of meridional
    mode numbers given in a list.
    :param he:
        Equivalent Depth
    :param n:
        Meridional Mode Number
    :param latitude:
        Latitude
    :param max_wn:
        Max global wave number.
        The global wave number range is (-max_wn,max_wn)
    :param n_wn:
        Number of global wave numbers in the range (-max_wn,max_wn)
    :type he: Float
    :type n: List of integers (e.g. [1,2,3])
    :type latitude: Float
    :type maxwn: Positive Integer (max_wn > 0)
    :type n_wn: Integer
    :return: DataFrame with wn and frequency
    :rtype: DataFrame
    """
    df = []
    df.append(kelvin_mode(he,latitude,max_wn,n_wn))
    df.append(mrg_mode(he,latitude,max_wn,n_wn))
    df.append(eig_n_0(he,latitude,max_wn,n_wn))

    for nn in n:
        df.append(er_n(he,nn,latitude,max_wn,n_wn))
        df.append(eig_n(he,nn,latitude,max_wn,n_wn))
        df.append(wig_n(he,nn,latitude,max_wn,n_wn))

    df = reduce(lambda left,right: pd.merge(left,right,on='Wavenumber'), df)
    return df

def standar_plot(he,size=12,figsize=(8, 8),mx_wn=20,mx_freq=1.,labels='on'):
    """
    Creates a standard plot with all Matsuno modes for a given Equivalent Depth
    he.  This function plots the dispersion curves for the meridional mode
    numbers n = [1,2,3]. Some configuration is possible although with several
    limitations.
    :param he:
        Equivalent Depth
    :param size(optional):
        Text size
    :param figsize(optional):
        Figure size
    :param mx_wn(optional):
        Upper Limit of the Zonal Wave number range. The range is given
        by (-mx_wn,mx_wn)
    :param mx_freq(optional):
        Upper Limit for the y axis given as frequency in CPD.
    :param labels(optional):
        Plot the name of the Matsuno Modes in the figureself.
    :type he: Float
    :type size: Integer
    :type figsize: Tuple
    :type mx_wn: Integer
    :type mx_freq: Float
    :type mx_freq: String
    :return: Plot with Matsuno Modes
    :rtype: Matplotlib Figure
    """

    plt.rc('font', size=size)          # controls default text sizes
    plt.rc('axes', titlesize=size)     # fontsize of the axes title
    plt.rc('axes', labelsize=size)    # fontsize of the x and y labels
    plt.rc('xtick', labelsize=size)    # fontsize of the tick labels
    plt.rc('ytick', labelsize=size)    # fontsize of the tick labels
    plt.rc('legend', fontsize=size)    # legend fontsize
    plt.rc('figure', titlesize=size)  # fontsize of the figure title

    df = matsuno_dataframe(he,n=[1,2,3])
    wn = df.index.values
    fig,ax = plt.subplots(figsize=figsize)

    for column in df:
        ax.plot(wn,df[column].values,color='k')

    ax.set_xlim(-mx_wn,mx_wn)
    ax.set_ylim(0,mx_freq)
    ax.set_xlabel('ZONAL WAVENUMBER')
    ax.set_ylabel('FREQUENCY (CPD)')
    plt.text(mx_wn-2*0.25*mx_wn,-0.06,'EASTWARD',fontsize=size-2)
    plt.text(-mx_wn+0.25*mx_wn,-0.06,'WESTWARD',fontsize=size-2)

    if labels=='on':
        # Print Kelvin Label
        p_wn = wn[np.logical_and(wn>=-mx_wn,wn<=mx_wn)]
        i = int((len(p_wn)/2)+0.3*(len(p_wn)/2))
        i, = np.where(wn == p_wn[i])[0]
        plt.text(wn[i]-1,df.iloc[i][0],'Kelvin', \
        bbox={'facecolor':'white','edgecolor':'none'},fontsize=size+1)

        # Print MRG Label
        p_wn = wn[np.logical_and(wn>=-mx_wn,wn<=mx_wn)]
        i = int(0.7*(len(p_wn)/2))
        i, = np.where(wn == p_wn[i])[0]
        plt.text(wn[i]-1,df.iloc[i][1],'MRG', \
        bbox={'facecolor':'white','edgecolor':'none'},fontsize=size+1)

        # Print EIG(n=0) Label
        p_wn = wn[np.logical_and(wn>=-mx_wn,wn<=mx_wn)]
        i = int((len(p_wn)/2)+0.1*(len(p_wn)/2))
        i, = np.where(wn == p_wn[i])[0]
        plt.text(wn[i]-1,df.iloc[i][2],'EIG(n=0)', \
        bbox={'facecolor':'white','edgecolor':'none'},fontsize=size+1)

        # Print ER Label
        p_wn = wn[np.logical_and(wn>=-mx_wn,wn<=mx_wn)]
        i = int(0.7*(len(p_wn)/2))
        i, = np.where(wn == p_wn[i])[0]
        plt.text(wn[i]-1,df.iloc[i][3]+0.01,'ER', \
        bbox={'facecolor':'none','edgecolor':'none'},fontsize=size+1)

        # Print EIG Label
        p_wn = wn[np.logical_and(wn>=-mx_wn,wn<=mx_wn)]
        i = int((len(p_wn)/2)+0.3*(len(p_wn)/2))
        i, = np.where(wn == p_wn[i])[0]
        plt.text(wn[i]-1,df.iloc[i][7],'EIG', \
        bbox={'facecolor':'white','edgecolor':'none'},fontsize=size+1)

        # Print WIG Label
        p_wn = wn[np.logical_and(wn>=-mx_wn,wn<=mx_wn)]
        i = int(0.55*(len(p_wn)/2))
        i, = np.where(wn == p_wn[i])[0]
        plt.text(wn[i]-1,df.iloc[i][8],'WIG', \
        bbox={'facecolor':'white','edgecolor':'none'},fontsize=size+1)

        # Print n Labels
        p_wn = wn[wn>=0]
        i, = np.where(wn == p_wn[0])[0]
        plt.text(-1,df.iloc[i][4],'n=1', \
        bbox={'facecolor':'white','edgecolor':'none'},fontsize=size+1)
        plt.text(-1,df.iloc[i][7],'n=2', \
        bbox={'facecolor':'white','edgecolor':'none'},fontsize=size+1)
        plt.text(-1,df.iloc[i][10],'n=3', \
        bbox={'facecolor':'white','edgecolor':'none'},fontsize=size+1)

    plt.show()
    return fig

def matsuno_modes_wk(he=[12,25,50],n=[1,],latitude=0.,max_wn=20,n_wn=500):
    """
    Creates a dataframe with all Matsuno modes for a given set of meridional
    mode numbers given in a list.
    :param he:
        Equivalent Depth
    :param n:
        Meridional Mode Number
    :param latitude:
        Latitude
    :param max_wn:
        Max global wave number.
        The global wave number range is (-max_wn,max_wn)
    :param n_wn:
        Number of global wave numbers in the range (-max_wn,max_wn)
    :type he: Float
    :type n: List of integers (e.g. [1,2,3])
    :type latitude: Float
    :type maxwn: Positive Integer (max_wn > 0)
    :type n_wn: Integer
    :return: DataFrame with wn and frequency
    :rtype: DataFrame
    """
    matsuno_modes = {}

    for h in he:
        df = []
        df.append(kelvin_mode(h,latitude,max_wn,n_wn))
        df.append(mrg_mode(h,latitude,max_wn,n_wn))
        df.append(eig_n_0(h,latitude,max_wn,n_wn))

        for nn in n:
            df.append(er_n(h,nn,latitude,max_wn,n_wn))
            df.append(eig_n(h,nn,latitude,max_wn,n_wn))
            df.append(wig_n(h,nn,latitude,max_wn,n_wn))

        df = reduce(lambda left,right: pd.merge(left,right,on='Wavenumber'), df)
        matsuno_modes[h] = df
    return matsuno_modes



import pandas as pd
import os

from metcalcpy import GROUP_SEPARATOR
from metcalcpy.agg_stat import AggStat


def test_groups():
    # prepare parameters
    params = {'random_seed': 1, 'indy_var': 'fcst_lead',
              'method': 'perc',
              'num_iterations': 1, 'event_equal': 'True',
              'derived_series_1': [
                  ['Group_y1_1 TMP ME', 'Group_y1_2 TMP ME', 'DIFF']],
              'derived_series_2': [],
              'agg_stat_input': 'data/agg_stat_with_groups_data.data',
              'fcst_var_val_1': {'TMP': ['ME']},
              'fcst_var_val_2': {},
              'agg_stat_output': 'data/agg_stat_with_groups_output.data',
              'fixed_vars_vals_input': {'fcst_lev': {'fcst_lev_0': ['Z02']}},
              'series_val_1': {'model': ['GTS+RAIN3mm' + GROUP_SEPARATOR + 'GTS+RAIN4mm',
                                         'GTS+RAIN50p' + GROUP_SEPARATOR + 'GTS+RAIN5mm']},
              'series_val_2': {},
              'alpha': 0.05, 'line_type': 'sl1l2',
              'num_threads': -1,
              'indy_vals': ['0', '120000', '240000'],
              'list_stat_1': ['ME'],
              'list_stat_2': []}
    # start aggregation logic
    AGG_STAT = AggStat(params)
    AGG_STAT.calculate_stats_and_ci()

    # read the output
    input_data = pd.read_csv(
        params['agg_stat_output'],
        header=[0],
        sep='\t'
    )
    assert len(input_data) == 9
    assert input_data.loc[(input_data['model'] == 'GTS+RAIN3mm' + GROUP_SEPARATOR + 'GTS+RAIN4mm')
                          & (input_data['fcst_lead'] == 120000)]['stat_value'].item() == -1.4384707
    assert input_data.loc[(input_data['model'] == 'GTS+RAIN50p' + GROUP_SEPARATOR + 'GTS+RAIN5mm')
                          & (input_data['fcst_lead'] == 240000)]['stat_value'].item() == -1.5292608
    assert input_data.loc[(input_data['model'] == 'DIFF(Group_y1_1 TMP ME-Group_y1_2 TMP ME)')
                          & (input_data['fcst_lead'] == 0)]['stat_value'].item() == -0.02839
    # remove the output
    os.remove(params['agg_stat_output'])



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
from metcalcpy.util.mode_ratio_statistics import *
from metcalcpy.util.utils import column_data_by_name_value, THREE_D_DATA_FILTER


def calculate_3d_ratio_fsa_asa(input_data, columns_names):
    """Performs calculation of  % of 3d simple objects that are forecast
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_fsa_asa(filtered_data, columns_names)


def calculate_3d_ratio_osa_asa(input_data, columns_names):
    """Performs calculation of % of 3d simple objects that are observation

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_osa_asa(filtered_data, columns_names)


def calculate_3d_ratio_asm_asa(input_data, columns_names):
    """Performs calculation of % of 3d simple objects that are matched

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_asm_asa(filtered_data, columns_names)


def calculate_3d_ratio_asu_asa(input_data, columns_names):
    """Performs calculation of % of 3d simple objects that are unmatched

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_asu_asa(filtered_data, columns_names)


def calculate_3d_ratio_fsm_fsa(input_data, columns_names):
    """Performs calculation of % of 3d simple forecast objects that are matched

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_fsm_fsa(filtered_data, columns_names)


def calculate_3d_ratio_fsu_fsa(input_data, columns_names):
    """Performs calculation of % of 3d simple forecast objects that are unmatched

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_fsu_fsa(filtered_data, columns_names)


def calculate_3d_ratio_osm_osa(input_data, columns_names):
    """Performs calculation of % of 3d simple simple observation objects that are matched

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_osm_osa(filtered_data, columns_names)


def calculate_3d_ratio_osu_osa(input_data, columns_names):
    """Performs calculation of % of 3d simple simple observation objects that are unmatched

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_osu_osa(filtered_data, columns_names)


def calculate_3d_ratio_fsm_asm(input_data, columns_names):
    """Performs calculation of % of 3d simple matched objects that are forecasts

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_fsm_asm(filtered_data, columns_names)


def calculate_3d_ratio_osm_asm(input_data, columns_names):
    """Performs calculation of % of 3d simple matched objects that are observations

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_osm_asm(filtered_data, columns_names)


def calculate_3d_ratio_fsu_asu(input_data, columns_names):
    """Performs calculation of % of 3d simple unmatched objects that are forecast

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_fsu_asu(filtered_data, columns_names)


def calculate_3d_ratio_osu_asu(input_data, columns_names):
    """Performs calculation of % of 3d simple unmatched objects that are observation

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_osu_asu(filtered_data, columns_names)


def calculate_3d_ratio_fsa_aaa(input_data, columns_names):
    """Performs calculation of ?

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_fsa_aaa(filtered_data, columns_names)


def calculate_3d_ratio_osa_aaa(input_data, columns_names):
    """Performs calculation of ?

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_osa_aaa(filtered_data, columns_names)


def calculate_3d_ratio_fsa_faa(input_data, columns_names):
    """Performs calculation of % of all 3d forecast objects that are simple

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_fsa_faa(filtered_data, columns_names)


def calculate_3d_ratio_fca_faa(input_data, columns_names):
    """Performs calculation of % of all 3d forecast objects that are cluster

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_fca_faa(filtered_data, columns_names)


def calculate_3d_ratio_osa_oaa(input_data, columns_names):
    """Performs calculation of % of all 3d observation objects that are simple

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_osa_oaa(filtered_data, columns_names)


def calculate_3d_ratio_oca_oaa(input_data, columns_names):
    """Performs calculation of % of all 3d observation objects that are cluster

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_oca_oaa(filtered_data, columns_names)


def calculate_3d_ratio_fca_aca(input_data, columns_names):
    """Performs calculation of % of 3d cluster objects that are forecast

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_fca_aca(filtered_data, columns_names)


def calculate_3d_ratio_fsa_osa(input_data, columns_names):
    """Performs calculation of Ratio of simple 3d forecasts to simple 3d observations [frequency bias]

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_fsa_osa(filtered_data, columns_names)


def calculate_3d_ratio_osa_fsa(input_data, columns_names):
    """Performs calculation of Ratio of simple 3d observations to simple 3d forecasts [1 / frequency bias]

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_osa_fsa(filtered_data, columns_names)


def calculate_3d_ratio_asa_aca(input_data, columns_names):
    """Performs calculation of Ratio of simple 3d objects to 3d cluster objects

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_asa_aca(filtered_data, columns_names)


def calculate_3d_ratio_fca_fsa(input_data, columns_names):
    """Performs calculation of Ratio of 3d cluster forecast objects to 3d simple forecast objects'

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_fca_fsa(filtered_data, columns_names)


def calculate_3d_ratio_fsa_fca(input_data, columns_names):
    """Performs calculation of Ratio of simple 3d forecast objects to cluster 3d forecast objects

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_fsa_fca(filtered_data, columns_names)


def calculate_3d_ratio_oca_osa(input_data, columns_names):
    """Performs calculation of Ratio of cluster 3d observation objects to simple 3d observation objects

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_oca_osa(filtered_data, columns_names)


def calculate_3d_ratio_osa_oca(input_data, columns_names):
    """Performs calculation of Ratio of simple 3d observation objects to cluster 3d observation objects

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_ratio_osa_oca(filtered_data, columns_names)


def calculate_3d_objhits(input_data, columns_names):
    """Performs calculation of Hits =/2

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_objhits(filtered_data, columns_names)


def calculate_3d_objmisses(input_data, columns_names):
    """Performs calculation of Misses = OSU

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_objmisses(filtered_data, columns_names)


def calculate_3d_objfas(input_data, columns_names):
    """Performs calculation of False Alarms = FSU

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_objfas(filtered_data, columns_names)


def calculate_3d_objcsi(input_data, columns_names):
    """Performs calculation of critical success index CSI = hits //2 + OSU + FSU]

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_objcsi(filtered_data, columns_names)


def calculate_3d_objpody(input_data, columns_names):
    """Performs calculation of Probability of Detecting Yes PODY = hits //2 + OSU]

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_objpody(filtered_data, columns_names)


def calculate_3d_objfar(input_data, columns_names):
    """Performs calculation of False alarm ratio FAR = false alarms //2 + FSU]

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated statistic as float
            or None if some of the data values are missing or invalid
    """
    filtered_data = column_data_by_name_value(input_data, columns_names, THREE_D_DATA_FILTER)
    return calculate_objfar(filtered_data, columns_names)
